import gfapy
import unittest

class TestLineVersion(unittest.TestCase):

  def test_header(self):
    self.assertEqual("generic", gfapy.Line("H\tVN:Z:1.0").version)
    self.assertEqual("gfa1", gfapy.Line("H\tVN:Z:1.0", version = "gfa1").version)
    self.assertEqual("gfa2", gfapy.Line("H\tVN:Z:1.0", version = "gfa2").version)

  def test_comment(self):
    self.assertEqual("generic", gfapy.Line("# VN:Z:1.0").version)
    self.assertEqual("gfa1", gfapy.Line("# VN:Z:1.0", version = "gfa1").version)
    self.assertEqual("gfa2", gfapy.Line("# VN:Z:1.0", version = "gfa2").version)

  def test_segment(self):
    self.assertEqual("gfa1", gfapy.Line("S\tA\tNNNN").version)
    self.assertEqual("gfa2", gfapy.Line("S\tA\t1\tNNNN").version)
    self.assertEqual("gfa1", gfapy.Line("S\tA\tNNNN", version = "gfa1").version)
    self.assertEqual("gfa2", gfapy.Line("S\tA\t1\tNNNN", version = "gfa2").version)
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line("S\tA\t1\tNNNN", version = "gfa1")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line("S\tA\tNNNN", version = "gfa2")

  def test_link(self):
    self.assertEqual("gfa1", gfapy.Line("L\tA\t+\tB\t-\t*").version)
    self.assertEqual("gfa1",
           gfapy.Line("L\tA\t+\tB\t-\t*", version = "gfa1").version)
    with self.assertRaises(gfapy.VersionError):
      gfapy.Line("L\tA\t+\tB\t-\t*", version = "gfa2")
    with self.assertRaises(gfapy.VersionError):
      gfapy.line.edge.Link(["A","+","B","-","*"], version = "gfa2")

  def test_containment(self):
    self.assertEqual("gfa1", gfapy.Line("C\tA\t+\tB\t-\t10\t*").version)
    self.assertEqual("gfa1",
       gfapy.Line("C\tA\t+\tB\t-\t10\t*", version = "gfa1").version)
    with self.assertRaises(gfapy.VersionError):
      gfapy.Line("C\tA\t+\tB\t-\t10\t*", version = "gfa2")
    with self.assertRaises(gfapy.VersionError):
      gfapy.line.edge.Containment(["A","+","B","-","10","*"], version = "gfa2")

  def test_custom_record(self):
    self.assertEqual("gfa2", gfapy.Line("X\tVN:Z:1.0").version)
    self.assertEqual("gfa2", gfapy.Line("X\tVN:Z:1.0", version = "gfa2").version)
    with self.assertRaises(gfapy.VersionError):
      gfapy.Line("X\tVN:Z:1.0", version = "gfa1")
    with self.assertRaises(gfapy.VersionError):
      gfapy.line.CustomRecord(["X","VN:Z:1.0"], version = "gfa1")



class AlignmentType:

  def is_internal(self):
    """Does the edge represent an internal alignment?

    Note:
      only GFA2 E lines may represent internal alignments

    Returns:
      bool
    """
    return self._alignment_type == "I"

  def is_containment(self):
    """Does the edge represent a containment?

    Note:
      A containment is either a C line (GFA1) or an E line, for which the
      coordinates of at least one of the two sequences go from 0 to the end of
      the sequence (GFA2).

    Returns:
      bool
    """
    return self._alignment_type == "C"

  def is_dovetail(self):
    """Does the edge represent a dovetail overlap?

    Note:
      A dovetail is either a L line (GFA1) or an E line (GFA2), for which the
      coordinates of both sequences go from the beginning of the sequence
      to some internal position, or from some internal position to the end of
      the sequence.

    Returns:
      bool
    """
    return self._alignment_type == "L"



class Writer:
  def __str__(self):
    return "#" + str(self.spacer) + str(self.content)

  def to_list(self):
    """Convert the content of the comment line to a list.

    The generic to_list() method of Line is overwritten,
    in order to support an optional spacer specification.
    """
    return ["#", self.content, self.spacer]



from ..common.alignment_type import AlignmentType
from ..common.from_to import FromTo
from ..gfa1.to_gfa2 import ToGFA2 as GFA1_ToGFA2
from ..gfa1.references import References as GFA1_References
from ..gfa1.oriented_segments import OrientedSegments
from ..gfa1.alignment_type import AlignmentType as GFA1_AlignmentType
from ..gfa1.other import Other
from .canonical import Canonical
from .complement import Complement
from .equivalence import Equivalence
from .references import References as Link_References
from .to_gfa2 import ToGFA2 as Link_ToGFA2
from ..edge import Edge

class Link(Link_ToGFA2, GFA1_ToGFA2, Link_References, Equivalence, Complement, \
    Canonical, Other, GFA1_AlignmentType, OrientedSegments, GFA1_References, \
    AlignmentType, FromTo, Edge):
  """A link line (L) of a GFA1 file.

  Note:
    from_segment and to_segment are used instead of from/to
    as from is not a valid method name in Python. However, when not
    used as method name (e.g. as argument of get()), from and to can
    be used, as an alias has been defined.
  """

  RECORD_TYPE = "L"
  POSFIELDS = ["from_segment", "from_orient", "to_segment", "to_orient",
               "overlap"]
  PREDEFINED_TAGS = ["MQ", "NM", "RC", "FC", "KC", "ID"]
  FIELD_ALIAS = {"from": "from_segment", "to": "to_segment"}
  DATATYPE = {
    "from_segment" : "segment_name_gfa1",
    "from_orient" : "orientation",
    "to_segment" : "segment_name_gfa1",
    "to_orient" : "orientation",
    "overlap" : "alignment_gfa1",
    "MQ" : "i",
    "NM" : "i",
    "RC" : "i",
    "FC" : "i",
    "KC" : "i",
    "ID" : "Z",
  }
  NAME_FIELD = "ID"
  REFERENCE_FIELDS = ["from_segment", "to_segment"]
  BACKREFERENCE_RELATED_FIELDS = ["to_orient", "from_orient", "overlap"]
  DEPENDENT_LINES = ["paths"]

Link._apply_definitions()



import gfapy
import unittest

class TestAPIReferencesGroups(unittest.TestCase):

  def test_paths_references(self):
    g = gfapy.Gfa()
    s = {}; l = {}
    for name in ["a", "b", "c", "d", "e", "f"]:
      s[name] = gfapy.Line("S\t{}\t*".format(name))
      g.append(s[name])
    path = gfapy.Line("P\tp1\tf+,a+,b+,c-,e+\t*")
    self.assertEqual([gfapy.OrientedLine("f","+"), gfapy.OrientedLine("a","+"),
                      gfapy.OrientedLine("b","+"), gfapy.OrientedLine("c","-"),
                      gfapy.OrientedLine("e","+")], path.segment_names)
    self.assertEqual([], path.links)
    # connection
    g.append(path)
    # add links
    for name in ["a+b+", "b+c-", "c-d+", "e-c+", "a-f-"]:
      l[name] = gfapy.Line("\t".join((list("L{}*".format(name)))))
      g.append(l[name])
    # segment_names
    self.assertEqual([gfapy.OrientedLine(s["f"],"+"),
                      gfapy.OrientedLine(s["a"],"+"),
                      gfapy.OrientedLine(s["b"],"+"),
                      gfapy.OrientedLine(s["c"],"-"),
                      gfapy.OrientedLine(s["e"],"+")], path.segment_names)
    # links
    self.assertEqual([gfapy.OrientedLine(l["a-f-"],"-"),
                      gfapy.OrientedLine(l["a+b+"],"+"),
                      gfapy.OrientedLine(l["b+c-"],"+"),
                      gfapy.OrientedLine(l["e-c+"],"-")],
                      path.links)
    # path disconnection
    path.disconnect()
    self.assertEqual([gfapy.OrientedLine("f","+"),
                      gfapy.OrientedLine("a","+"),
                      gfapy.OrientedLine("b","+"),
                      gfapy.OrientedLine("c","-"),
                      gfapy.OrientedLine("e","+")], path.segment_names)
    self.assertEqual([], path.links)
    g.append(path)
    # links disconnection cascades on paths:
    assert(path.is_connected())
    l["a-f-"].disconnect()
    assert(not path.is_connected())
    self.assertEqual([gfapy.OrientedLine("f","+"),
                      gfapy.OrientedLine("a","+"),
                      gfapy.OrientedLine("b","+"),
                      gfapy.OrientedLine("c","-"),
                      gfapy.OrientedLine("e","+")], path.segment_names)
    g.append(path)
    g.append(l["a-f-"])
    # segment disconnection cascades on links and then paths:
    assert(path.is_connected())
    s["a"].disconnect()
    assert(not path.is_connected())
    self.assertEqual([gfapy.OrientedLine("f","+"),
                      gfapy.OrientedLine("a","+"),
                      gfapy.OrientedLine("b","+"),
                      gfapy.OrientedLine("c","-"),
                      gfapy.OrientedLine("e","+")], path.segment_names)
    self.assertEqual([], path.links)

  def test_paths_backreferences(self):
    g = gfapy.Gfa()
    s = {}; l = {}
    for name in ["a", "b", "c", "d", "e", "f"]:
      s[name] = gfapy.Line("S\t{}\t*".format(name))
      g.append(s[name])
    path = gfapy.Line("P\tp1\tf+,a+,b+,c-,e+\t*")
    g.append(path)
    for sname in ["a", "b", "c", "e", "f"]:
      self.assertEqual([path], s[sname].paths)
    self.assertEqual([], s["d"].paths)
    for name in ["a+b+", "b+c-", "c-d+", "e-c+", "a-f-"]:
      l[name] = gfapy.Line("\t".join(list("L{}*".format(name))))
      g.append(l[name])
    for lname in ["a+b+", "b+c-", "e-c+", "a-f-"]:
      self.assertEqual([path], l[lname].paths)
    self.assertEqual([], l["c-d+"].paths)
    # disconnection effects
    path.disconnect()
    for lname in ["a+b+", "b+c-", "c-d+", "e-c+", "a-f-"]:
      self.assertEqual([], l[lname].paths)
    for sname in ["a", "b", "c", "d", "e", "f"]:
      self.assertEqual([], s[sname].paths)
    # reconnection
    path.connect(g)
    for sname in ["a", "b", "c", "e", "f"]:
      self.assertEqual([path], s[sname].paths)
    self.assertEqual([], s["d"].paths)
    for lname in ["a+b+", "b+c-", "e-c+", "a-f-"]:
      self.assertEqual([path], l[lname].paths)
    self.assertEqual([], l["c-d+"].paths)

  def test_gfa2_paths_references(self):
    g = gfapy.Gfa()
    s = {}
    for name in ["a", "b", "c", "d", "e", "f"]:
      s[name] = gfapy.Line("S\t{}\t1000\t*".format(name))
      g.append(s[name])
    path1_part1 = gfapy.Line("O\tp1\tp2- b+")
    path1_part2 = gfapy.Line("O\tp1\tc- e-c+-")
    path1 = path1_part2
    path2 = gfapy.Line("O\tp2\tf+ a+")
    self.assertEqual([gfapy.OrientedLine("p2","-"),
                      gfapy.OrientedLine("b","+")], path1_part1.items)
    self.assertEqual([gfapy.OrientedLine("c","-"),
                      gfapy.OrientedLine("e-c+","-")], path1_part2.items)
    self.assertEqual([gfapy.OrientedLine("f","+"),
                      gfapy.OrientedLine("a","+")], path2.items)
    with self.assertRaises(gfapy.RuntimeError): path1.captured_path
    with self.assertRaises(gfapy.RuntimeError): path2.captured_path
    # connection
    g.append(path1_part1)
    g.append(path1_part2)
    g.append(path2)
    # edges
    e = {}
    for name in ["a+b+", "b+c-", "c-d+", "e-c+", "a-f-", "f-b+"]:
      coord1 = "900\t1000$" if (name[1] == "+") else "0\t100"
      coord2 = "0\t100" if (name[3] == "+")  else "900\t1000$"
      e[name] = gfapy.Line("E\t{}\t{}\t{}\t{}\t{}\t100M".format(name,name[0:2],name[2:4],coord1,coord2))
      g.append(e[name])
    # items
    self.assertEqual([gfapy.OrientedLine(path2,"-"),
                      gfapy.OrientedLine(s["b"],"+"),
                      gfapy.OrientedLine(s["c"],"-"),
                      gfapy.OrientedLine(e["e-c+"],"-")],
                      path1.items)
    self.assertEqual([gfapy.OrientedLine(s["f"],"+"),
                      gfapy.OrientedLine(s["a"],"+")],
                      path2.items)
    # induced set
    self.assertEqual([gfapy.OrientedLine(s["f"],"+"),
                      gfapy.OrientedLine(e["a-f-"],"-"),
                      gfapy.OrientedLine(s["a"],"+")],
                      path2.captured_path)
    self.assertEqual([gfapy.OrientedLine(s["a"],"-"),
                      gfapy.OrientedLine(e["a-f-"],"+"),
                      gfapy.OrientedLine(s["f"],"-"),
                      gfapy.OrientedLine(e["f-b+"],"+"),
                      gfapy.OrientedLine(s["b"],"+"),
                      gfapy.OrientedLine(e["b+c-"],"+"),
                      gfapy.OrientedLine(s["c"],"-"),
                      gfapy.OrientedLine(e["e-c+"],"-"),
                      gfapy.OrientedLine(s["e"],"+")],
                      path1.captured_path)
    # backreferences
    for line in [path2, s["b"], s["c"], e["e-c+"]]:
      self.assertEqual([path1], line.paths)
    for line in [s["f"], s["a"]]:
      self.assertEqual([path2], line.paths)
    # group disconnection
    path1.disconnect()
    self.assertEqual([gfapy.OrientedLine("p2","-"), gfapy.OrientedLine("b","+"), gfapy.OrientedLine("c","-"), gfapy.OrientedLine("e-c+","-")],
                 path1.items)
    with self.assertRaises(gfapy.RuntimeError):
      path1.captured_path
    self.assertEqual([gfapy.OrientedLine(s["f"],"+"), gfapy.OrientedLine(s["a"],"+")], path2.items)
    for line in [path2, s["b"], s["c"], e["e-c+"]]:
      self.assertEqual([], line.paths)
    # group reconnection
    g.append(path1)
    self.assertEqual([gfapy.OrientedLine(path2,"-"), gfapy.OrientedLine(s["b"],"+"), gfapy.OrientedLine(s["c"],"-"), gfapy.OrientedLine(e["e-c+"],"-")],
                 path1.items)
    self.assertEqual([gfapy.OrientedLine(s["f"],"+"), gfapy.OrientedLine(s["a"],"+")], path2.items)
    for line in [path2, s["b"], s["c"], e["e-c+"]]:
      self.assertEqual([path1], line.paths)
    # item disconnection cascades on group
    assert(path1.is_connected())
    assert(path2.is_connected())
    e["e-c+"].disconnect()
    assert(not path1.is_connected())
    assert(path2.is_connected())
    g.append(e["e-c+"])
    g.append(path1)
    # two-level disconnection cascade
    assert(path1.is_connected())
    assert(path2.is_connected())
    s["f"].disconnect()
    assert(not path2.is_connected())
    assert(not path1.is_connected())

  def test_sets_references(self):
    g = gfapy.Gfa()
    s = {}
    set1 = gfapy.Line("U\tset1\tb set2 c e-c+")
    set2 = gfapy.Line("U\tset2\tg c-d+ path1")
    path1 = gfapy.Line("O\tpath1\tf+ a+")
    self.assertEqual(["b", "set2", "c", "e-c+"], set1.items)
    self.assertEqual(["g", "c-d+", "path1"], set2.items)
    # induced set of non-connected cannot be computed
    with self.assertRaises(gfapy.RuntimeError): set1.induced_set
    with self.assertRaises(gfapy.RuntimeError): set2.induced_set
    # connection
    g.append(set1)
    g.append(set2)
    # induced set cannot be computed, as long as not all references are solved
    with self.assertRaises(gfapy.RuntimeError): set1.induced_set
    # connect items
    g.append(path1)
    for name in ["a", "b", "c", "d", "e", "f", "g"]:
      s[name] = gfapy.Line("S\t"+"{}".format(name)+"\t1000\t*")
      g.append(s[name])
    e = {}
    for name in ["a+b+", "b+c-", "c-d+", "e-c+", "a-f-"]:
      coord1 = "900\t1000$" if (name[1] == "+") else "0\t100"
      coord2 = "0\t100" if (name[3] == "+")  else "900\t1000$"
      e[name] = gfapy.Line("E\t{}\t{}\t{}\t{}\t{}\t100M".format(name,name[0:2],name[2:4],coord1,coord2))
      g.append(e[name])
    # items
    self.assertEqual([s["b"], set2, s["c"], e["e-c+"]], set1.items)
    self.assertEqual([s["g"], e["c-d+"], path1], set2.items)
    # induced set
    self.assertEqual([gfapy.OrientedLine(s["f"],"+"), gfapy.OrientedLine(s["a"],"+")],
                 path1.captured_segments)
    self.assertEqual(set([x.name for x in [s["g"], s["c"], s["d"], s["f"], s["a"]]]),
                     set([x.name for x in set2.induced_segments_set]))
    self.assertEqual(set([x.name for x in [s["b"], s["g"], s["c"], s["d"], s["f"], s["a"], s["e"]]]),
                     set([x.name for x in set1.induced_segments_set]))
    self.assertEqual(set([x.name for x in [e["c-d+"], e["a-f-"]]]),
                     set([x.name for x in set2.induced_edges_set]))
    self.assertEqual([e["a+b+"],e["b+c-"],e["c-d+"],e["e-c+"],e["a-f-"]],
                 set1.induced_edges_set)
    self.assertEqual(set([x.name for x in set1.induced_segments_set + set1.induced_edges_set]),
                     set([x.name for x in set1.induced_set]))
    # backreferences
    for line in [s["b"], set2, s["c"], e["e-c+"]]: 
      self.assertEqual([set1], line.sets)
    for line in [s["g"], e["c-d+"], path1]: 
      self.assertEqual([set2], line.sets)
    # group disconnection
    set1.disconnect()
    self.assertEqual(["b", "set2", "c", "e-c+"], set1.items)
    for line in [s["b"], set2, s["c"], e["e-c+"]]: 
      self.assertEqual([], line.sets)
    # group reconnection
    g.append(set1)
    self.assertEqual([s["b"], set2, s["c"], e["e-c+"]], set1.items)
    for line in [s["b"], set2, s["c"], e["e-c+"]]: 
      self.assertEqual([set1], line.sets)
    # item disconnection cascades on group
    assert(set1.is_connected())
    e["e-c+"].disconnect()
    assert(not set1.is_connected())
    g.append(e["e-c+"])
    g.append(set1)
    # multilevel disconnection cascade
    assert(path1.is_connected())
    assert(set2.is_connected())
    assert(set1.is_connected())
    s["f"].disconnect()
    assert(not path1.is_connected())
    assert(not set2.is_connected())
    assert(not set1.is_connected())




import gfapy

class RedundantLinearPaths:

  def _junction_junction_paths(self, sn, exclude):
    retval = []
    exclude.add(sn)
    s = self.segment(sn)
    for dL in s.dovetails_L:
      eL = dL.other_end(gfapy.SegmentEnd(s, "L"))
      if (eL.name in exclude) or (len(eL.segment.dovetails_of_end(eL.end_type)) == 1):
        retval.append([True, eL, gfapy.SegmentEnd(s, "R"), True])
    for dR in s.dovetails_R:
      eR = dR.other_end(gfapy.SegmentEnd(s, "R"))
      if (eR.name in exclude) or (len(eR.segment.dovetails_of_end(eR.end_type)) == 1):
        retval.append([True, gfapy.SegmentEnd(s, "R"), eR.inverted(), True])
    return retval

  def _extend_linear_path_to_junctions(self, segpath):
    segfirst = self.segment(segpath[0].segment)
    segfirst_d = segfirst.dovetails_of_end(gfapy.invert(segpath[0].end_type))
    redundant_first = (len(segfirst_d) > 0)
    if len(segfirst_d) == 1:
      segpath.insert(0, segfirst_d[0].other_end(segpath[0].inverted()))
    segpath.insert(0, redundant_first)
    seglast = self.segment(segpath[-1].segment)
    seglast_d = seglast.dovetails_of_end(segpath[-1].end_type)
    redundant_last = (len(seglast_d) > 0)
    if len(seglast_d) == 1:
      segpath.append(seglast_d[0].other_end(segpath[-1].inverted()))
    segpath.append(redundant_last)

  def _link_duplicated_first(self, merged, first, is_reversed, jntag):
    # annotate junction
    if jntag is None:
      jntag = "jn"
    if not first.get(jntag):
      first.set(jntag, {"L":[],"R":[]})
    if is_reversed:
      first.get(jntag)["L"].append([merged.name, "-"])
    else:
      first.get(jntag)["R"].append([merged.name, "+"])
    # create temporary link
    ln = len(first.sequence)
    if self._version == "gfa1":
      tmp_link = gfapy.line.edge.Link([first.name, \
        "-" if is_reversed else "+", merged.name, "+", \
        "{}M".format(ln), "co:Z:temporary"])
      self.add_line(tmp_link)
    elif self._version == "gfa2":
      tmp_link = gfapy.line.edge.GFA2(["*",first.name + \
        ("-" if is_reversed else "+"), merged.name+"+",
        "0" if is_reversed else str(ln-1), # on purpose fake
        "1" if is_reversed else "{}$".format(ln), # on purpose fake
        0, str(ln), "{}M".format(ln), "co:Z:temporary"])
      self.add_line(tmp_link)
    else:
      raise gfapy.AssertionError()

  def _link_duplicated_last(self, merged, last, is_reversed, jntag):
    # annotate junction
    if jntag is None:
      jntag = "jn"
    if not last.get(jntag):
      last.set(jntag, {"L":[],"R":[]})
    if is_reversed:
      last.get(jntag)["R"].append([merged.name, "-"])
    else:
      last.get(jntag)["L"].append([merged.name, "+"])
    # create temporary link
    ln = len(last.sequence)
    if self._version == "gfa1":
      tmp_link = gfapy.line.edge.Link([merged.name, "+",
          last.name, "-" if is_reversed else "+",
          "{}M".format(ln), "co:Z:temporary"])
      self.add_line(tmp_link)
    elif self._version == "gfa2":
      mln = len(merged.sequence)
      tmp_link = gfapy.line.edge.GFA2(["*",merged.name+"+", \
        last.name+("-" if is_reversed else "+"),
        str(mln - ln), "{}$".format(mln),
        str(ln-1) if is_reversed else "0", # on purpose fake
        "{}$".format(ln) if is_reversed else "1", # on purpose fake
        "{}M".format(ln), "co:Z:temporary"])
      self.add_line(tmp_link)
    else:
      raise gfapy.AssertionError()

  def _remove_junctions(self, jntag):
    if jntag is None:
      jntag = "jn"
    for s in self.segments:
      jndata = s.get(jntag)
      if jndata:
        ln = len(s.sequence)
        for m1, dir1 in jndata["L"].items():
          for m2, dir2 in jndata["R"].items():
            if self._version == "gfa1":
              l = gfapy.line.edge.Link([m1,dir1,m2,dir2,"{}M".format(ln)])
              self.add_line(l)
            elif self._version == "gfa2":
              m1ln = len(self.segment(m1).sequence)
              m2ln = len(self.segment(m2).sequence)
              r1 = (dir1 == "-")
              r2 = (dir2 == "-")
              l = gfapy.line.edge.GFA2(["*", m1+dir1, m2+dir2,
                 "0" if r1 else str(m1ln-ln),
                 str(ln) if r1 else str(m1ln)+"$",
                 "0" if r2 else str(m2ln-ln),
                 str(ln) if r1 else str(m2ln)+"$",
                 str(ln)+"M"])
              self.add_line(l)
            else:
              raise gfapy.AssertionError()
        s.disconnect()




import gfapy
import gfapy.line
from .collections import Collections
from .headers import Headers
from .creators import Creators
from .destructors import Destructors
from .finders import Finders

class Lines(Collections, Creators, Destructors, Finders, Headers):

  GFA1Specific = [
                   gfapy.line.edge.Link,
                   gfapy.line.edge.Containment,
                   gfapy.line.group.Path,
                   gfapy.line.segment.GFA1
                 ]

  GFA2Specific = [
                   gfapy.line.CustomRecord,
                   gfapy.line.Fragment,
                   gfapy.line.Gap,
                   gfapy.line.edge.GFA2,
                   gfapy.line.segment.GFA2,
                   gfapy.line.group.Unordered,
                   gfapy.line.group.Ordered,
                   gfapy.line.Unknown
                 ]

  def _api_private_check_gfa_line(self, gfa_line, callermeth):
    if not isinstance(gfa_line, gfapy.Line):
      raise gfapy.TypeError("Note: {} is API private, ".format(callermeth)+
          "do not call it directly\n"+
          "Error: line class is {} and not gfapy.Line")
    elif not gfa_line._gfa is self:
      raise gfapy.RuntimeError("Note: {} is API private, ".format(callermeth)+
          "do not call it directly\n"+
          "Error: line.gfa is not the expected instance of gfapy.Gfa\n"+
          repr(gfa_line.gfa)+" != "+repr(self))



import gfapy

class FieldData:
  """
  Disallow editing the VN tag in connected header lines
  """

  def _set_existing_field(self, fieldname, value, set_reference=False):
    if fieldname == "VN" and self.get("VN") is not None and self.is_connected():
      raise gfapy.RuntimeError(
        "The value of the header tag VN cannot be edited\n"+
        "For version conversion use to_gfa1 or to_gfa2")
    else:
      super()._set_existing_field(fieldname, value,
                                  set_reference=set_reference)



import gfapy
import re

def unsafe_decode(string):
  return gfapy.NumericArray.from_string(string, valid = True)

def decode(string):
  return gfapy.NumericArray.from_string(string)

def validate_encoded(string):
  if not re.match(r"^(f(,[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)+|[CSI](,\+?[0-9]+)+|[csi](,[-+]?[0-9]+)+)$", string):
    raise gfapy.FormatError(
      "{} is not a valid numeric array string\n".format(repr(string))+
      "(it must be one of [fcsiCSI] followed by a comma-separated list of:"+
      " for f: floats; for csi: signed integers; for CSI: unsigned integers)")

def validate_decoded(numeric_array):
  numeric_array.validate()

def unsafe_encode(obj):
  if isinstance(obj, gfapy.NumericArray):
    return str(obj)
  elif isinstance(obj, list):
    return str(gfapy.NumericArray(obj))
  elif isinstance(obj, str):
    return obj
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, line, gfapy.NumericArray)")

def encode(obj):
  if isinstance(obj, gfapy.NumericArray):
    return str(obj)
  elif isinstance(obj, list):
    return str(gfapy.NumericArray(obj))
  elif isinstance(obj, str):
    validate_encoded(obj)
    return obj
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, line, gfapy.NumericArray)")



import gfapy

class Equivalence:

  def __hash__(self):
    name = self.get("name")
    if name:
      return name.__hash__()
    else:
      return NotImplemented

  def __eq__(self, o):
    """
    Equivalence check

    Returns
    -------
    bool
      does the line has the same record type,
      contains the same tags
      and all positional fields and tags contain the same field values?

    See Also
    --------
    gfapy.line.edge.Link.__eq__
    """
    if o is self:
      return True
    if isinstance(o, str):
      name = self.get("name")
      if name:
        return name == str(o)
    if not isinstance(o, gfapy.Line):
      return False
    if (o.record_type != self.record_type):
      return False
    if sorted(o._data.keys()) != sorted(self._data.keys()):
      return False
    for k,v in o._data.items():
      if self._data[k] != v:
        if self.field_to_s(k) != o.field_to_s(k):
          return False
    return True

  def diff(self, other):
    if self.record_type != other.record_type:
      return ("incompatible", "record_type", \
              self.record_type, other.record_type)
    if self.__class__ != other.__class__:
      if self.version == other.version:
        raise gfapy.AssertionError
      return ("incompatible", "version", self.version, other.version)
    differences = []
    for fieldname in self.positional_fieldnames:
      value1 = self.field_to_s(fieldname)
      value2 = other.field_to_s(fieldname)
      if value1 != value2:
        differences.append(("different", "positional_field",
                            fieldname, value1, value2))
    for tagname in other.tagnames:
      if tagname not in self.tagnames:
        differences.append(("exclusive", ">", "tag",
                      tagname, other.get_datatype(tagname),
                      other.get(tagname)))
    for tagname in self.tagnames:
      if tagname not in other.tagnames:
        differences.append(("exclusive", "<", "tag",
                      tagname, self.get_datatype(tagname), self.get(tagname)))
      else:
        tag1 = self.field_to_s(tagname, tag=True)
        tag2 = other.field_to_s(tagname, tag=True)
        if tag1 != tag2:
          differences.append(("different", "tag", tagname,
                          self.get_datatype(tagname),
                          self.field_to_s(tagname),
                          other.get_datatype(tagname),
                          other.field_to_s(tagname)))
    return differences

  def diffscript(self, other, selfvar):
    outscript = []
    for diffitem in self.diff(other):
      if diffitem[0] == "incompatible":
        if diffitem[1] == "record_type":
          raise gfapy.RuntimeError(
            "Cannot compute conversion script: different record type\n"+
            "Line: {}\n".format(self)+
            "Other: {}\n".format(other)+
            "{0} != {1}",format(diffitem[2], diffitem[3]))
        elif diffitem[1] == "version":
          raise gfapy.RuntimeError(
            "Cannot compute conversion script: different GFA version\n"+
            "Line: {}\n".format(self)+
            "Other: {}\n".format(other)+
            "{0} != {1}",format(diffitem[2], diffitem[3]))
      elif diffitem[0] == "different":
        if diffitem[1] == "positional_field":
          outscript.append("{0}.set('{1}', '{2}')".format(selfvar,
                                      diffitem[2].replace("'","\\'"),
                                      diffitem[4].replace("'","\\'")))
        elif diffitem[1] == "tag":
          if diffitem[3] != diffitem[5]:
            outscript.append("{0}.set_datatype('{1}', '{2}')".format(selfvar,
                                      diffitem[2].replace("'","\\'"),
                                      diffitem[5].replace("'","\\'")))
          if diffitem[4] != diffitem[6]:
            outscript.append("{0}.set('{1}', '{2}')".format(selfvar,
                                      diffitem[2].replace("'","\\'"),
                                      diffitem[6].replace("'","\\'")))
      elif diffitem[0] == "exclusive":
        if diffitem[1] == ">":
          if diffitem[2] == "tag":
            outscript.append("{0}.set_datatype('{1}', '{2}')".format(selfvar,
                                      diffitem[3].replace("'","\\'"),
                                      diffitem[4].replace("'","\\'")))
            outscript.append("{0}.set('{1}', '{2}')".format(selfvar,
                                      diffitem[3].replace("'","\\'"),
                                      diffitem[5].replace("'","\\'")))
        elif diffitem[1] == "<":
          if diffitem[2] == "tag":
            outscript.append("{0}.delete('{1}')".format(selfvar,
                                      diffitem[3].replace("'","\\'")))
    return "\n".join(outscript)


  def _has_field_values(self, hsh, ignore_fields = None):
    assert(isinstance(hsh, dict))
    if ignore_fields is None:
      ignore_fields = []
    if ("record_type" in hsh) and ("record_type" not in ignore_fields) \
        and (self.record_type != hsh["record_type"]):
      return False
    ignore_fields.append("record_type")
    fieldnames = [i for i in hsh.keys() if i not in ignore_fields]
    for fieldname in fieldnames:
      value = self.get(fieldname)
      if value is None:
        return False
      if gfapy.is_placeholder(value):
        continue
      if value != hsh[fieldname] and \
          (self.field_to_s(fieldname) != hsh[fieldname]):
        return False
    return True

  def _has_eql_fields(self, refline, ignore_fields = None):
    assert(isinstance(refline, gfapy.Line))
    if ignore_fields is None:
      ignore_fields = []
    self._dealias_fieldnames(ignore_fields)
    if ("record_type" not in ignore_fields) \
        and (self.record_type != refline.record_type):
      return False
    fieldnames = refline.positional_fieldnames + refline.tagnames
    fieldnames = [i for i in fieldnames if i not in ignore_fields]
    if "name" in ignore_fields:
      name_field = refline.__class__.NAME_FIELD
      if name_field in fieldnames:
        fieldnames.remove(name_field)
    for fieldname in fieldnames:
      refvalue = refline.get(fieldname)
      if gfapy.is_placeholder(refvalue):
        continue
      value = self.get(fieldname)
      if value is None:
        return False
      if gfapy.is_placeholder(value):
        continue
      if value != refvalue:
        return False
    return True



import gfapy

class CapturedPath:

  @property
  def captured_segments(self):
    return [ x for x in self.captured_path if isinstance(x.line, gfapy.line.segment.GFA2) ]

  @property
  def captured_edges(self):
    return [ x for x in self.captured_path if isinstance(x.line, gfapy.line.edge.GFA2) ]

  @property
  def captured_path(self):
    if not self.is_connected():
      raise gfapy.RuntimeError(
        "Captured path cannot be computed\n"+
        "Line is not connected to a GFA instance\n"+
        "Line: {}".format(self))
    return self._compute_captured_path()[0]

  def _compute_captured_path(self):
    path = []
    prev_edge = False
    for item in self.items:
      path, prev_edge = self._push_item_on_se_path(path, prev_edge, item)
    return path, prev_edge

  def _push_item_on_se_path(self, path, prev_edge, item):
    if isinstance(item.line, str):
      raise gfapy.RuntimeError(
        "Captured path cannot be computed; a reference has not been resolved\n"+
        "Line: {}\n".format(self)+
        "Unresolved reference: {} (String found)".format(item.line))
    elif isinstance(item.line, gfapy.line.segment.GFA2):
      if not item.line.is_connected():
        raise gfapy.RuntimeError(
          "Captured path cannot be computed; item is not connected\n"+
          "Line: {}\n".format(self)+
          "Item: {}".format(item.line))
      self._push_segment_on_se_path(path, prev_edge, item)
      prev_edge = False
    elif isinstance(item.line, gfapy.line.edge.GFA2):
      if not item.line.is_connected():
        raise gfapy.RuntimeError(
          "Captured path cannot be computed; item is not connected\n"+
          "Line: {}\n".format(self)+
          "Item: {}".format(item.line))
      if not path:
        self._push_first_edge_on_se_path(path, self.items)
      else:
        self._push_nonfirst_edge_on_se_path(path, item)
      prev_edge = True
    elif isinstance(item.line, gfapy.line.group.Ordered):
      if not item.line.is_connected():
        raise gfapy.RuntimeError(
          "Captured path cannot be computed; item is not connected\n"+
          "Line: {}\n".format(self)+
          "Item: {}".format(item.line))
      subpath, prev_edge_subpath = item.line._compute_captured_path()
      if not subpath:
        raise gfapy.AssertionError()
      if item.orient == "+":
        for subpath_item in subpath:
          path, prev_edge = self._push_item_on_se_path(path, prev_edge,
              subpath_item)
      else:
        for subpath_item in reversed(subpath):
          path, prev_edge = self._push_item_on_se_path(path, prev_edge,
              subpath_item.inverted())
      prev_edge = prev_edge_subpath
    elif isinstance(item.line, gfapy.line.unknown.Unknown):
      raise gfapy.RuntimeError(
        "Captured path cannot be computed; a reference has not been resolved\n"+
        "Line: {}\n".format(self)+
        "Unresolved reference: {} (Virtual unknown line)".format(item.name))
    else:
      raise gfapy.TypeError(
        "Line: {}\t".format(self)+
        "Cannot compute captured path:\t"+
        "Error: items of type {} are not supported\t".format(item.line.__class__.__name__)+
        "Unsupported item: {}".format(item))
    return path, prev_edge

  def _push_first_edge_on_se_path(self, path, items):
    oriented_edge = items[0]
    oss = [oriented_edge.line.sid1, oriented_edge.line.sid2]
    if oriented_edge.orient == "-":
      for i in range(len(oss)):
        oss[i] = oss[i].inverted()
    if len(items) > 1:
      nextitem = items[1]
      if isinstance(nextitem.line, gfapy.line.segment.GFA2):
        if nextitem == oss[0]:
          oss.reverse()
        # if oss does not include nextitem an error will be raised
        # in the next iteration, so does not need to be handled here
      elif isinstance(nextitem.line, gfapy.line.edge.GFA2):
        oss_of_next = [nextitem.line.sid1, nextitem.line.sid2]
        if oriented_edge.orient == "-":
          for i in range(len(oss_of_next)):
            oss_of_next[i] = oss_of_next[i].inverted()
        if oss[0] in oss_of_next:
          oss.reverse()
        # if oss_of_next have no element in common with oss an error will be
        # raised in the next iteration, so does not need to be handled here
      elif isinstance(nextitem.line, gfapy.line.group.Ordered):
        subpath = nextitem.line.captured_path
        if not subpath: return# does not need to be further handled here
        if nextitem.orient == "+":
          firstsubpathsegment = subpath[0]
        else:
          firstsubpathsegment = subpath[-1].inverted()
        if firstsubpathsegment == oss[0]:
          oss.reverse()
        # if oss does not include in firstsubpathsegment
        # error will be raised in next iteration, ie not handled here
      else:
        pass
        # don't need to handle here other cases, as they will be handled
        # in the next iteration of push_item_on_se_path
    path.append(oss[0])
    path.append(oriented_edge)
    path.append(oss[1])

  def _push_nonfirst_edge_on_se_path(self, path, oriented_edge):
    prev_os = path[-1]
    path.append(oriented_edge)
    possible_prev = [oriented_edge.line.sid1, oriented_edge.line.sid2]
    if oriented_edge.orient == "-":
      for i, v in enumerate(possible_prev):
        possible_prev[i] = possible_prev[i].inverted()
    if prev_os == possible_prev[0]:
      path.append(possible_prev[1])
    elif prev_os == possible_prev[1]:
      path.append(possible_prev[0])
    else:
      raise gfapy.NotFoundError(
        "Path is not valid, elements are not contiguous\n"+
        "Line: {}\n".format(self)+
        "Previous elements:\n"+
        "".join(["  {} ({})\n".format(e, e.line) for e in path])+
        "Current element:\n"+
        "  {} ({})".format(oriented_edge, oriented_edge.line))

  def _push_segment_on_se_path(self, path, prev_edge, oriented_segment):
    if path:
      if isinstance(path[-1].line, gfapy.line.segment.GFA2):
        if prev_edge:
          self._check_s_is_as_expected(path, oriented_segment)
          return # do not add segment, as it is already there
        else:
          path.append(self._find_edge_from_path_to_segment(path, oriented_segment))
      elif isinstance(path[-1].line, gfapy.line.edge.GFA2):
        self._check_s_to_e_contiguity(path, oriented_segment)
      else:
        raise gfapy.AssertionError()
    path.append(oriented_segment)

  def _check_s_is_as_expected(self, path, oriented_segment):
    if path[-1] != oriented_segment:
      raise gfapy.InconsistencyError(
        "Path is not valid\n"+
        "Line: {}\n".format(self)+
        "Previous elements:\n"+
        "".join(["  {} ({})\n".format(e, e.line) for e in path[0:-2]])+
        "Expected element:\n"+
        "  {} ({})\n".format(path[-1], path[-1].line)+
        "Current element:\n"+
        "  {} ({})\n".format(oriented_segment, oriented_segment.line))

  def _check_s_to_e_contiguity(self, path, oriented_segment):
    # check that segment is an extremity of path[-1]
    # and that the other extremity is path[-2]
    if not (path[-1].sid1 == self.segment and path[-1].sid2 == path[-2]) and \
       not (path[-1].sid1 == path[-2] and path[-1].sid2 == self.segment):
      raise gfapy.InconsistencyError(
        "Path is not valid\n"+
        "Line: {}\n".format(self)+
        "Previous elements:\n"+
        "".join(["  {} ({})\n".format(e, e.line) for e in path])+
        "Current element:\n"+
        "  {} ({})\n".format(oriented_segment, oriented_segment.line))

  def _find_edge_from_path_to_segment(self, path, oriented_segment):
    edges = []
    for edge in oriented_segment.line.edges:
      if (edge.sid1 == oriented_segment and edge.sid2 == path[-1]) or \
         (edge.sid1 == path[-1] and edge.sid2 == oriented_segment):
        edges.append(gfapy.OrientedLine(edge, "+"))
      elif (edge.sid1 == oriented_segment.inverted() and
            edge.sid2 == path[-1].inverted()) or\
           (edge.sid1 == path[-1].inverted() and
            edge.sid2 == oriented_segment.inverted()):
        edges.append(gfapy.OrientedLine(edge, "-"))
    if len(edges) == 0:
      raise gfapy.NotFoundError(
        "Path is not valid, segments are not contiguous\n"+
        "Line: {}\n".format(self)+
        "Previous elements:\n"+
        "".join(["  {} ({})\n".format(e, e.line) for e in path])+
        "Current element:\n"+
        "  {} ({})\n".format(oriented_segment, oriented_segment.line))
    elif len(edges) > 1:
      raise gfapy.NotUniqueError(
        "Path is not unique\n"+
        "Line: {}\n".format(self)+
        "Previous elements:\n"+
        "".join(["  {} ({})\n".format(e, e.line) for e in path])+
        "Current element:\n"+
        "  {} ({})\n".format(oriented_segment, oriented_segment.line)+
        "Possible edges\n"+
        "".join(["  {} ({})\n".format(e, e.line) for e in edges]))
    return edges[0]

  def _check_captured_path_elem_connected(self, item):
    if not item.is_connected():
      raise gfapy.RuntimeError(
        "Cannot compute induced set\n"+
        "Non-connected element found\n"+
        "Item: {}\nLine: {}".format(item, self))



import gfapy
import unittest
from .extension import *

class TestAPIExtensions(unittest.TestCase):

  def test_extensions(self):
    g = gfapy.Gfa(version="gfa2", vlevel=0)
    MetagenomicAssignment(["M", "*","N12","C","SC:i:20"])
    sA = gfapy.Line("S\tA\t1000\t*")
    g.append(sA)
    tB12 = gfapy.Line("T\tB12_c")
    g.append(tB12)
    m1 = gfapy.Line("M\t1\ttaxon:123\tA\tSC:i:40\txx:Z:cjaks536")
    g.append(m1)
    m2 = gfapy.Line("M\t2\ttaxon:123\tB\txx:Z:cga5r5cs")
    g.append(m2)
    sB = gfapy.Line("S\tB\t1000\t*")
    g.append(sB)
    mx = gfapy.Line("M\t*\tB12_c\tB\tSC:i:20")
    g.append(mx)
    t123 = gfapy.Line(
      "T\ttaxon:123\tUL:Z:http://www.taxon123.com")
    g.append(t123)
    self.assertEqual(MetagenomicAssignment, m1.__class__)
    self.assertEqual(Taxon, tB12.__class__)
    self.assertEqual("1", m1.mid)
    assert(gfapy.is_placeholder(mx.mid))
    self.assertEqual(t123, m1.tid)
    self.assertEqual(sA, m1.sid)
    self.assertEqual("cjaks536", m1.xx)
    self.assertEqual([m2,mx], sB.metagenomic_assignments)
    self.assertEqual([m1,m2], t123.metagenomic_assignments)
    self.assertEqual("taxon:123", t123.tid)
    self.assertEqual("http://www.taxon123.com", t123.UL)




from .gfa2_to_gfa1 import GFA2ToGFA1
from .coverage import Coverage
from .references import References
from .writer_wo_sequence import WriterWoSequence
from . import Segment

class GFA2(WriterWoSequence, References, Coverage, GFA2ToGFA1, Segment):
  """A segment line of a GFA file"""

  VERSION = "gfa2"
  RECORD_TYPE = "S"
  POSFIELDS = ["sid", "slen", "sequence"]
  PREDEFINED_TAGS = ["RC", "FC", "KC", "SH", "UR"]
  DATATYPE = {
    "sid" : "identifier_gfa2",
    "slen" : "i",
    "sequence" : "sequence_gfa2",
    "RC" : "i",
    "FC" : "i",
    "KC" : "i",
    "SH" : "H",
    "UR" : "Z",
  }
  NAME_FIELD = "sid"
  FIELD_ALIAS = { "name" : "sid", "length" : "slen", "LN" : "slen" }
  DEPENDENT_LINES = ["dovetails_L", "dovetails_R", "gaps_L", "gaps_R",
                     "edges_to_contained", "edges_to_containers",
                     "fragments", "internals", "paths", "sets"]

GFA2._apply_definitions()



import json
from copy import deepcopy

class Cloning:

  def clone(self):
    """Copy of a gfapy.Line instance.
    The copy will be disconnected, ie do not belong to the GFA and do not
    contain cross-references to other lines. This allows to edit the line
    (eg. changing the unique ID) before adding it.
    To achieve this, all reference fields are copied in their string
    representation.
    All other fields are copied as they are, and a deep copy is done for
    arrays, strings and JSON fields.

    Returns
    -------
    gfapy.Line
    """
    data_cpy = {}
    for k,v in self._data.items():
      if k in self.__class__.REFERENCE_FIELDS:
        data_cpy[k] = self.field_to_s(k)
      elif self._field_datatype(k) == "J":
        data_cpy[k] = json.loads(json.dumps(v))
      elif isinstance(v, list) or isinstance(v, str):
        data_cpy[k] = deepcopy(v)
      else:
        data_cpy[k] = v
    cpy = self.__class__(data_cpy, vlevel = self.vlevel,
                         virtual = self.virtual, version = self.version)
    cpy._datatype = self._datatype.copy()
    # cpy._refs and cpy._gfa are not set, so that the cpy is disconnected
    return cpy



import gfapy

class CapturedPath:

  @property
  def captured_edges(self):
    if not self.is_connected():
      raise gfapy.RuntimeError(
        "Captured path cannot be computed\n"+
        "Line is not connected to a GFA instance\n"+
        "Line: {}".format(self))
    return self.links

  @property
  def captured_segments(self):
    if not self.is_connected():
      raise gfapy.RuntimeError(
        "Captured path cannot be computed\n"+
        "Line is not connected to a GFA instance\n"+
        "Line: {}".format(self))
    return self.segment_names

  @property
  def captured_path(self):
    if not self.is_connected():
      raise gfapy.RuntimeError(
        "Captured path cannot be computed\n"+
        "Line is not connected to a GFA instance\n"+
        "Line: {}".format(self))
    retval = []
    for i in range(len(self.segment_names) - 1):
      retval.append(self.segment_names[i])
      retval.append(self.links[i])
    retval.append(self.segment_names[-1])
    if len(self.segment_names) == len(self.links):
      retval.append(self.links[-1])
      retval.append(self.segment_names[0])
    return retval



from itertools import groupby
import gfapy

class InvertibleSegments:

  def randomly_orient_invertibles(self):
    ''' Selects a random orientation for all invertible segments.

    For the definition of invertible segment, see Gonnella and Kurtz (2016).'''
    for sn in self.segment_names:
      if self._segment_same_links_both_ends(sn):
        self._randomly_orient_proven_invertible_segment(sn)

  def randomly_orient_invertible(self, segment):
    '''Selects a random orientation for an invertible segment.

    For the definition of invertible segment, see Gonnella and Kurtz (2016).'''
    if isinstance(segment, gfapy.Line):
      segment_name = segment.name
    else:
      segment_name = segment
    if not self._segment_same_links_both_ends(segment_name):
      raise gfapy.RuntimeError("Only segments with links to the same or "+
          "equivalent segments at both ends can be randomly oriented")
    self._randomly_orient_proven_invertible_segment(segment_name)

  def _randomly_orient_proven_invertible_segment(self, segment_name):
    se = gfapy.SegmentEnd([segment_name, "R"])
    parts = self._partitioned_links_of(se)
    if len(parts) == 2:
      tokeep1_other_end = parts[0][0].other_end(se)
      tokeep2_other_end = parts[1][0].other_end(se)
    elif len(parts) == 1 and len(parts[0]) == 2:
      tokeep1_other_end = parts[0][0].other_end(se)
      tokeep2_other_end = parts[0][1].other_end(se)
    else:
      return
    if len(tokeep1_other_end.segment.dovetails(
          tokeep1_other_end.end_type)) < 2:
      return
    if len(tokeep2_other_end.segment.dovetails(
          tokeep2_other_end.end_type)) < 2:
      return
    self._delete_other_links(se, tokeep1_other_end)
    self._delete_other_links(se.inverted(), tokeep2_other_end)
    self._annotate_random_orientation(segment_name)

  def _link_targets_for_cmp(segment_end):
    return ["".join(str(l.other_end(segment_end))) \
            for l in segment_end.segment.dovetails(segment_end.end_type)]

  def _segment_same_links_both_ends(self, segment_name):
    e_links = self._link_targets_for_cmp(gfapy.SegmentEnd(segment_name, "R"))
    b_links = self._link_targets_for_cmp(gfapy.SegmentEnd(segment_name, "L"))
    return e_links == b_links

  def _segment_signature(self, segment_end):
    s = self.try_get_segment(segment_end.segment)
    return ",".join(self._link_targets_for_cmp(segment_end))+"\t"+\
           ",".join(self._link_targets_for_cmp(segment_end.inverted()))+"\t"+\
           s.field_to_s("or")

  def _partitioned_links_of(self, segment_end):
    links = segment_end.segment.dovetails(segment_end.end_type)
    sigs = {}
    for l in links:
      sigs[id(l)] = self._segment_signature(l.other_end(segment_end))
    sig = lambda l: sigs[id(l)]
    return [list(v) for k,v in groupby(sorted(links,key=sig),key=sig)]

  def _annotate_random_orientation(self, segment_name):
    segment = self.try_get_segment(segment_name)
    n = segment.name.split("_")
    pairs = 0
    pos = [1, segment.LN]
    if segment.get("or"):
      o = segment.field_to_s("or").split(",")
      if len(o) > 2:
        while o[-1]=="{}^".format(o[0]) or o[0]=="{}^".format(o[-1]):
          pairs += 1
          o.pop()
          o.pop(0)
      if segment.mp:
        pos = [segment.mp[pairs*2], segment.mp[-1-pairs*2]]
    rn = segment.rn
    if rn is None:
      rn = []
    rn += pos
    segment.rn = rn
    n[pairs] = "({}".format(n[pairs])
    n[-1-pairs] = "{})".format(n[-1-pairs])
    self.rename(segment.name, "_".join(n))



import gfapy

class Validation:
  def _validate_lists_size(self):
    n_overlaps = len(self.overlaps)
    n_segments = len(self.segment_names)
    if n_overlaps == n_segments - 1:
      # case 1: linear path
      return True
    elif n_overlaps == 1 and not self.overlaps[0]:
      # case 2: linear path, single "*" to represent overlaps which are all "*"
      return True
    elif n_overlaps == n_segments:
      # case 3: circular path
      pass
    else:
      raise gfapy.InconsistencyError(
        "Path has {} oriented segments, ".format(n_segments)+
        "but {} overlaps".format(n_overlaps))

  def _validate_record_type_specific_info(self):
    self._validate_lists_size()



import gfapy

class References:

  def _prepare_and_check_ref(self, ref):
    if isinstance(ref, str):
      ref = self._line_for_ref_symbol(ref)
    self._check_ref_class(ref)
    self._check_ref_connection(ref)
    self._check_ref_not_self(ref)
    return ref

  def _check_ref_class(self, item):
    if item.__class__ not in [
         gfapy.line.edge.GFA2,
         gfapy.line.segment.GFA2,
         gfapy.line.gap.Gap,
         gfapy.line.group.Ordered,
         self.__class__]:
      raise gfapy.ArgumentError(
        "Line: {}\n".format(self)+
        "Cannot add items of class {}\n".format(item.__class__.__name__)+
        "Only GFA2 edges, segments, gaps, groups[*] "+
        "can be added\n(* = unordered groups to unordered groups only).")

  def _check_ref_connection(self, item):
    if item.line.gfa != self._gfa:
      raise gfapy.ArgumentError(
        "Line: {}\n".format(self)+
        "Item: {}".format(repr(item))+
        "The item added to the group must be connected\n"+
        "to the same GFA object as the group")

  def _check_ref_not_self(self, item):
    if (item.line == self):
      raise gfapy.RuntimeError(
        "Line: {}\n".format(self)+
        "Item is the line itself\n"+
        "A group is not allowed to refer to itself")

  def _line_for_ref_symbol(self, ref):
    line = self._gfa.line(ref)
    if line is None:
      if self._gfa._segments_first_order:
        raise gfapy.NotFoundError("Group: {}\n".format(self)+
                        "requires a non-existing ref with ID {}".format(ref))
      line = gfapy.line.unknown.Unknown({"name" : ref}, virtual = True,
                                        version = "gfa2")
      self._gfa.add_line(line)
    line._add_reference(self, "paths" if (self.record_type == "O") else "sets")
    return line



import unittest
import gfapy

class TestApiComments(unittest.TestCase):

  def test_initialize(self):
   l = gfapy.line.Comment("# hallo")
   self.assertEqual("# hallo", str(l))
   l = gfapy.line.Comment(["#", "hallo", "\t"])
   self.assertEqual("#\thallo", str(l))

  def test_fields(self):
   l = gfapy.line.Comment("# hallo")
   self.assertEqual("hallo", l.content)
   self.assertEqual(" ", l.spacer)
   l.content = "hello"
   self.assertEqual("hello", l.content)
   self.assertEqual("# hello", str(l))
   l.spacer = "  "
   self.assertEqual("hello", l.content)
   self.assertEqual("#  hello", str(l))

  def test_validation(self):
   with self.assertRaises(gfapy.FormatError):
     gfapy.line.Comment(["#", "hallo\nhallo"])
   with self.assertRaises(gfapy.FormatError):
     gfapy.line.Comment(["#", "hallo", "\n"])
   gfapy.line.Comment(["#", "hallo", "\n"], vlevel=0) # nothing raised
   l = gfapy.line.Comment(["#", "hallo"])
   l.content = "hallo\n" # nothing raised
   with self.assertRaises(gfapy.FormatError): str(l)
   l.content = "hallo"
   str(l) # nothing raised
   l.spacer = "\n" # nothing raised
   with self.assertRaises(gfapy.FormatError): str(l)
   l = gfapy.line.Comment(["#", "hallo"], vlevel=3)
   with self.assertRaises(gfapy.FormatError): l.content = "hallo\n"
   with self.assertRaises(gfapy.FormatError): l.spacer = "\n"

  def test_from_string(self):
    s = "# this is a comment"
    l = gfapy.Line(s)
    self.assertEqual(gfapy.line.Comment, l.__class__)
    self.assertEqual(s[2:], l.content)
    self.assertEqual(" ", l.spacer)
    s = "#this is another comment"
    l = gfapy.Line(s)
    self.assertEqual(gfapy.line.Comment, l.__class__)
    self.assertEqual(s[1:], l.content)
    self.assertEqual("", l.spacer)
    s = "#\t and this too"
    l = gfapy.Line(s)
    self.assertEqual(gfapy.line.Comment, l.__class__)
    self.assertEqual(s[3:], l.content)
    self.assertEqual(s[1:3], l.spacer)
    s = "#: and this too"
    l = gfapy.Line(s)
    self.assertEqual(gfapy.line.Comment, l.__class__)
    self.assertEqual(s[1:], l.content)
    self.assertEqual("", l.spacer)

  def test_to_s(self):
    s = "# this is a comment"
    l = gfapy.Line(s)
    self.assertEqual(s, str(l))
    s = "#this is another\tcomment"
    l = gfapy.Line(s)
    self.assertEqual(s, str(l))
    s = "#this is another\tcomment"
    l = gfapy.Line(s)
    l.spacer = " "
    self.assertEqual("# "+s[1:], str(l))

  def test_tags(self):
    with self.assertRaises(gfapy.ValueError):
      gfapy.line.Comment(["#", "hallo", " ", "zz:Z:hallo"])
    l = gfapy.Line("# hallo zz:Z:hallo")
    self.assertEqual("hallo zz:Z:hallo", l.content)
    self.assertEqual(None, l.zz)
    with self.assertRaises(gfapy.RuntimeError): l.zz = 1
    with self.assertRaises(gfapy.RuntimeError): l.set("zz", 1)
    self.assertEqual(None, l.get("zz"))

  def test_to_gfa1(self):
    s = "# this is a comment"
    l = gfapy.Line(s,version="gfa2")
    self.assertEqual(gfapy.line.Comment, l.__class__)
    self.assertEqual("gfa2", l.version)
    self.assertEqual(s, str(l))
    self.assertEqual("gfa2", l.to_gfa2().version)
    self.assertEqual(s, str(l.to_gfa2()))
    self.assertEqual("gfa1", l.to_gfa1().version)
    self.assertEqual(s, str(l.to_gfa1()))

  def test_to_gfa2(self):
    s = "# this is a comment"
    l = gfapy.Line(s,version="gfa1")
    self.assertEqual(gfapy.line.Comment, l.__class__)
    self.assertEqual("gfa1", l.version)
    self.assertEqual(s, str(l))
    self.assertEqual("gfa1", l.to_gfa1().version)
    self.assertEqual(s, str(l.to_gfa1()))
    self.assertEqual("gfa2", l.to_gfa2().version)
    self.assertEqual(s, str(l.to_gfa2()))

  def test_rgfa_comments(self):
    gfa = gfapy.Gfa()
    c1 = "#this is a comment"
    c2 = "# this is also a comment"
    c3 = "#and \tthis too!"
    gfa.add_line(c1) # nothing raised
    gfa.add_line(c2) # nothing raised
    gfa.add_line(c3) # nothing raised
    self.assertEqual([c1,c2,c3], [str(x) for x in gfa.comments])
    self.assertEqual(c1, str(gfa.comments[0]))
    gfa.rm(gfa.comments[0])
    self.assertEqual([c2,c3], [str(x) for x in gfa.comments])
    gfa.comments[0].disconnect()
    self.assertEqual([c3], [str(x) for x in gfa.comments])



import gfapy
from types import MethodType

class DynamicField:
  def __init__(self, get, set):
    self.get = get
    self.set = set

class DynamicFields:
  """
  Methods are dynamically created for non-existing but valid tag names.
  Methods for predefined tags and positional fields
  are created dynamically for each subclass; methods for existing tags
  are created on instance initialization.
  """

  def __getattribute__(self, name):
    try:
      attr = super().__getattribute__(name)
      if not isinstance(attr, DynamicField):
        return attr
      else:
        return attr.get(self)
    except AttributeError as err:
      return self._get_dynamic_field(name, err)

  def __setattr__(self, name, value):
    try:
      attr = super().__getattribute__(name)
      if not isinstance(attr, DynamicField):
        return super().__setattr__(name, value)
      else:
        attr.set(self, value)
    except AttributeError:
      return self._set_dynamic_field(name, value)

  def _get_dynamic_field(self, name, err):
    if self.virtual:
      raise err
    if name.startswith("try_get_"):
      name = name[8:]
      try_get = True
    else:
      try_get = False
    if name in self._data:
      return (lambda : self.try_get(name)) if try_get else self.get(name)
    if (name in self.__class__.PREDEFINED_TAGS or
        self._is_valid_custom_tagname(name)):
      if not try_get:
        return None
      else:
        raise gfapy.NotFoundError(
          "No value defined for tag {}".format(name))
    else:
      raise err

  def _set_dynamic_field(self, name, value):
    try:
      virtual = super().__getattribute__("_virtual")
      data = super().__getattribute__("_data")
      if virtual:
        super().__setattr__(name, value)
      if name in data:
        self._set_existing_field(name, value)
      if (name in self.__class__.PREDEFINED_TAGS or
            self._is_valid_custom_tagname(name)):
        self.set(name, value)
      else:
        super().__setattr__(name, value)
    except AttributeError:
      super().__setattr__(name, value)

  def _define_field_methods(self, fieldname):
    """Define field methods for a single field"""
    def getter(self):
      return self.get(fieldname)
    def try_get(self):
      return self.try_get(fieldname)
    def setter(self, value):
      self._set_existing_field(fieldname, value)
    super().__setattr__(fieldname, DynamicField(getter, setter))
    super().__setattr__("try_get_" + fieldname, MethodType(try_get, self))

# XXX: class methods are missing



import gfapy

class AlignmentPlaceholder(gfapy.Placeholder):
  """
  A placeholder subclass for alignment fields.

  Instances are usually created from their string representations, using the
  :class:`~gfapy.alignment.alignment.Alignment` factory class constructor.
  """

  def complement(self):
    """For compatibility with CIGAR alignments
    Returns:
      AlignmentPlaceholder : self
    """
    return self

  def __repr__(self):
    return "gfapy.AlignmentPlaceholder()"



class SegmentEndsPath(list):
  """
  a list containing {gfapy.SegmentEnd} elements, which defines a path
  in the graph
  """

  def reverse(self):
    """
    Reverses the direction of the path in place
    """
    self[:] = list(reversed(self))

  def __reversed__(self):
    """
    Iterator over the reverse-direction path
    """
    for elem in SegmentEndsPath(reversed([segment_end.inverted()
                                for segment_end in self])):
      yield elem



import gfapy

class References:

  def append_item(self, item):
    """
    Add an item to the group as last item.

    Parameters
    ----------
    item : gfapy.Line or str
    	GFA2 edge, segment, gap or group line to add.
    """
    if not self.is_connected():
      self._add_item_to_unconnected_group(item, True)
    else:
      self._add_item_to_connected_group(item, True)
      self.compute_induced_set() # check contiguity

  def prepend_item(self, item):
    """
    Add an item to the group as first item.

    Parameters
    ----------
    item : gfapy.Line or str
    	GFA2 edge, segment, gap or group line to add.
    """
    if not self.is_connected():
      self._add_item_to_unconnected_group(item, False)
    else:
      self._add_item_to_connected_group(item, False)
      self.compute_induced_set() # check contiguity

  def rm_first_item(self):
    """
    Remove the first item from the group.

    Parameters
    ----------
    item : str or gfapy.Line
    	GFA2 edge, segment, gap or group line to remove.
    """
    if not self.is_connected():
      self.items = self.items[1:]
    else:
      self.items[0].update_reference(self, "paths")
      self._delete_reference(self.items[0], "items")
      self.compute_induced_set() # check contiguity

  def rm_last_item(self):
    """
    Remove the last item from the group.

    Parameters
    ----------
    item : str or gfapy.Line
    	GFA2 edge, segment, gap or group line to remove.
    """
    if not self.is_connected():
      self.items = self.items[0:-1]
    else:
      self.items[-1].update_reference(self, "paths")
      self._delete_reference(self.items[-1], "items")
      self.compute_induced_set() # check contiguity

  def _add_item_to_unconnected_group(self, item, append = True):
    if isinstance(item.line, gfapy.Line):
      item.line = item.name
    if append:
      self.items.append(item)
    else:
      self.items.insert(0, item)

  def _add_item_to_connected_group(self, item, append = True):
    item.line = self.prepare_and_check_ref(item.line)
    self._add_reference(item, "items", append = append)

  def _initialize_references(self):
    for i in range(len(self.items)):
      self.items[i].line = self._line_for_ref_symbol(self.items[i].line)



import gfapy

class FieldDatatype:

  def get_datatype(self, fieldname):
    """
    Returns a string, which specifies the datatype of a field.

    Parameters
    ----------
    fieldname : str
      The tag name of the field.

    Returns
    -------
    str
      The datatype symbol.
      (One of gfapy.Field.FIELD_DATATYPE)
    """
    fieldname = self.__class__.FIELD_ALIAS.get(fieldname, fieldname)
    return self._field_or_default_datatype(fieldname,
        self._data.get(fieldname,None))

  def set_datatype(self, fieldname, datatype):
    """
    Set the datatype of a tag.

    If an existing tag datatype is changed, its content may become
    invalid (call **validate_field** if necessary).

    Parameters
    ----------
    fieldname : str
      The field name (it is not required that the field exists already)
    datatype : gfapy.Field.FIELD_DATATYPE
      The datatype.

    Raises
    ------
    gfapy.ArgumentError
      If **datatype** is not a valid datatype for tags.
    """
    if self._is_predefined_tag(fieldname):
      if self.get_datatype(fieldname) != datatype:
        raise gfapy.RuntimeError(
          "Cannot set the datatype of {} to {}\n".format(fieldname, datatype)+
          "The datatype of a predefined tag cannot be changed")
    elif not self._is_valid_custom_tagname(fieldname) and self.vlevel > 0:
      raise gfapy.FormatError(
        "{} is not a valid custom tag name".format(fieldname))
    if datatype not in gfapy.Field.TAG_DATATYPE:
      raise gfapy.ArgumentError("Unknown datatype: {}".format(datatype))
    self._datatype[fieldname] = datatype

  def _field_datatype(self, fieldname):
    return self._datatype.get(fieldname,
        self.__class__.DATATYPE.get(fieldname, None))

  def _field_or_default_datatype(self, fieldname, value):
    t = self._field_datatype(fieldname)
    if t is None:
      if value is None:
        return None
      t = gfapy.Field._get_default_gfa_tag_datatype(value)
      self._datatype[fieldname] = t
    return t



import gfapy

class Pos:

  @property
  def rpos(self):
    """The rightmost coordinate of the contained sequence in the container.

    Returns:
      int : 0-based coordinate

    Raises:
      gfapy.ValueError : If the overlap is a placeholder, thus the computation
        cannot be performed.
    """
    if isinstance(self.overlap, gfapy.Placeholder):
      raise gfapy.ValueError("The overlap is a placeholder, therefore"+
          "rpos cannot be computed")
    return self.pos + self.overlap.length_on_reference()



import gfapy

class FieldArray:
  """Multiple values of the same tag in different header lines.

  Parameters:
    datatype (str): (one of `gfapy.field.Field.TAG_DATATYPE`) the datatype of
      the tags represented by the array.
    data (list, None): a list of values. The values must be compatible with the
      specified datatype. If no list is provided, the instance is initialized
      with an empty list.
  """

  def __init__(self, datatype, data = None):
    if data is None:
      self._data = []
    else:
      self._data = data
    self._datatype = datatype

  @property
  def datatype(self):
    """Datatype of the tags represented by the list.

    Returns:
      str : one of `gfapy.field.Field.TAG_DATATYPE`.
    """
    return self._datatype

  def validate(self, fieldname : str = None) -> None:
    """Datatype-specific validation on each element of the list.

    Parameters:
      fieldname (str) : optional, for error messages.
    """
    self._validate_gfa_field(None, fieldname)

  def __str__(self):
    return self._to_gfa_field(self)

  def __repr__(self):
    return "gfapy.FieldArray({},{})".format(
        repr(self._datatype),repr(self._data))

  def __eq__(self, other):
    if isinstance(other, list):
      return other == self._data
    elif isinstance(other, gfapy.FieldArray):
      return other.datatype == self._datatype and \
          other.data == self._data
    else:
      return False

  def __add__(self, other):
    if isinstance(other, list):
      self._data += other
    elif isinstance(other, gfapy.FieldArray):
      self._data += other._data

  def __iter__(self):
    return self._data.__iter__()

  def __getattr__(self, name):
    return getattr(self._data, name)

  def _validate_gfa_field(self, datatype : str, fieldname = None):
    """Datatype-specific validation.

    If no datatype is provided as parameter, the datatype of the
    array is used.
    """
    if not datatype:
      datatype = self._datatype
    for elem in self._data:
      gfapy.Field._validate_gfa_field(elem, datatype, fieldname)

  def _default_gfa_tag_datatype(self):
    """
    Default GFA tag datatype.

    Returns
    -------
    gfapy.Field::TAG_DATATYPE
    """
    return self.datatype

  def _to_gfa_field(self, datatype = None, fieldname = None):
    """Representation as tab-separated values (w/o XX:Y: prefixes)."""
    if datatype is None:
      datatype = self._datatype
    return "\t".join(
        [ gfapy.Field._to_gfa_field(x, datatype = self._datatype, \
             fieldname = fieldname) for x in self._data ])

  def _to_gfa_tag(self, fieldname, datatype = None):
    """Representation as tab-separated tags (XX:Y:VALUE)."""
    if datatype is None:
      datatype = self.datatype
    return "\t".join(
        [ gfapy.Field._to_gfa_tag(x, fieldname, datatype) \
            for x in self._data ])

  def _vpush(self, value, datatype=None, fieldname=None):
    """Add a value to the array and validate.

    Raises
    ------
    gfapy.InconsistencyError
    	If the type of the new value does not correspond to the type of
      existing values.

    Parameters
    ----------
    value : Object
    	The value to add.
    datatype : gfapy.Field.TAG_DATATYPE or None
    	The datatype to use.
      If not **None**, it will be checked that the specified datatype is the
      same as for previous elements of the field array.
      If **None**, the value will be validated, according to the datatype
      specified on field array creation.
    fieldname : str
    	The field name to use for error messages.
    """
    if datatype is None:
      gfapy.Field._validate_gfa_field(value, self.datatype, fieldname)
    elif datatype != self.datatype:
      raise gfapy.InconsistencyError(
        "Datadatatype mismatch error for field {}:\n".format(fieldname)+
        "value: {}\n".format(value)+
        "existing datatype: {};\n".format(self.datatype)+
        "new datatype: {}".format(datatype))
    self._data.append(value)




import gfapy

class AlignmentType:

  @property
  def _alignment_type(self):
    """The alignment type (C, L or I).

    Returns:
      C, L or I: C if containment, L if link, I (internal) otherwise.
    """
    st1 = self._substring_type(self.beg1, self.end1)[0]
    st2 = self._substring_type(self.beg2, self.end2)[0]
    return self._alignment_type_for_substring_types(st1, st2)

  def _alignment_type_for_substring_types(self, st1, st2):
    if st1 == "whole" or st2 == "whole":
      return "C"
    elif self.sid1.orient == self.sid2.orient:
      if (st1 == "pfx" and st2 == "sfx") or (st1 == "sfx" and st2 == "pfx"):
        return "L"
      else:
        return "I"
    else:
      if (st1 == "pfx" and st2 == "pfx") or (st1 == "sfx" and st2 == "sfx"):
        return "L"
      else:
        return "I"

  def _substring_type(self, begpos, endpos):
    """Type of substring (pfx, sfx, whole, internal) given start and end pos.

    Analyzes the begin and end position and determine if the substring is
    the whole string, or a (possibly empty) other substring, ie a prefix,
    a suffix, or an internal alignment.
    """
    if gfapy.posvalue(begpos) > gfapy.posvalue(endpos):
      raise gfapy.ValueError(
        "Line: {}\n".format(str(self))+
        "begin > end: {}$ > {}".format(gfapy.posvalue(begpos),
                                       gfapy.posvalue(endpos)))
    if gfapy.isfirstpos(begpos):
      if gfapy.isfirstpos(endpos):
        return ("pfx", True)
      elif gfapy.islastpos(endpos):
        return ("whole", False)
      else:
        return ("pfx", False)
    elif gfapy.islastpos(begpos):
      if not gfapy.islastpos(endpos):
        raise gfapy.FormatError(
          "Line: {}\n".format(str(self))+
          "Wrong use of $ marker\n"+
          "{} >= {}$".format(gfapy.posvalue(endpos),
                             gfapy.posvalue(begpos)))
      return ("sfx", True)
    else:
      if gfapy.islastpos(endpos):
        return ("sfx", False)
      else:
        return ("internal",
            gfapy.posvalue(begpos) == gfapy.posvalue(endpos))



import gfapy

class Construction:

  @property
  def positional_fieldnames(self):
    """The names of the positional fields.

    The property is implemented differently for CustomRecord
    instances, as the positional fieldnames are unknown.

    Returns:
      list of str
    """
    return self._positional_fieldnames

  @property
  def tagnames(self):
    """The names of the tags defined in the line.

    The property is implemented differently for CustomRecord
    instances, as tags are identified heuristically (as the number
    of positional fields is unknown).

    Returns:
      list of str
    """
    return[x for x in self._data.keys() \
             if (not x in self.positional_fieldnames) \
                 and (x != "record_type")]

  def _initialize_positional_fields(self, strings):
    """delayed, see #delayed_inizialize_positional_fields"""
    pass

  def _initialize_tags(self, strings):
    first_tag = len(strings)
    for i in range(len(strings)-1, 0, -1):
      try:
        self._initialize_tag(*(gfapy.Field._parse_gfa_tag(strings[i])))
      except:
        break
      first_tag = i
    self._delayed_initialize_positional_fields(strings, first_tag)

  def _delayed_initialize_positional_fields(self, strings, n_positional_fields):
    self._positional_fieldnames = []
    if strings[0] in ["P", "C", "L"]:
      raise gfapy.VersionError(
        "GFA-like line (P,C,L) found in GFA2\n"+
        "Line: {}\n".format(" ".join(strings))+
        "Custom lines with record_type P, C and L are not supported by gfapy.")
    self._init_field_value("record_type", "custom_record_type", strings[0],
                     errmsginfo = strings)
    for i in range(1, n_positional_fields):
      n = "field{}".format(i)
      self._init_field_value(n, "generic", strings[i], errmsginfo = strings)
      self.positional_fieldnames.append(n)
      self._datatype[n] = "generic"



import gfapy
import re

def unsafe_decode(string):
  return string

def decode(string):
  validate_encoded(string)
  return string

def validate_encoded(string):
  if not re.match("^[!-~]+$", string):
    raise gfapy.FormatError(
      "{} is not a valid GFA2 identifier\n".format(repr(string))+
      "(it contains spaces or non-printable characters)")

def validate_decoded(obj):
  if isinstance(obj, gfapy.Line):
    validate_encoded(obj.name)
  elif isinstance(obj, str):
    validate_encoded(obj)
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, gfapy.Line)")

def unsafe_encode(obj):
  if isinstance(obj, str):
    return obj
  if isinstance(obj, gfapy.Line):
    return str(obj.name)
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, gfapy.Line)")

def encode(obj):
  string = unsafe_encode(obj)
  validate_encoded(string)
  return string



import unittest
import gfapy

class TestCigar(unittest.TestCase):

  def test_from_string_nonempty(self):
    self.assertEqual(gfapy.CIGAR([gfapy.CIGAR.Operation(12, "M"),
                                  gfapy.CIGAR.Operation( 1, "D"),
                                  gfapy.CIGAR.Operation( 2, "I")]),
                     gfapy.CIGAR._from_string("12M1D2I"))

  def test_from_string_empty(self):
    self.assertEqual(gfapy.Placeholder, gfapy.CIGAR._from_string("*"))

  def test_from_string_invalid(self):
    with self.assertRaises(gfapy.FormatError):
      gfapy.CIGAR._from_string("12x1D2I")

  def test__str__noempty(self):
    self.assertEqual("12M1D2I",
                     str(gfapy.CIGAR([gfapy.CIGAR.Operation(12, "M"),
                                      gfapy.CIGAR.Operation( 1, "D"),
                                      gfapy.CIGAR.Operation( 2, "I")])))



import gfapy
import re

def unsafe_decode(string):
  try:
    return int(string)
  except:
    raise gfapy.FormatError(
      "{} does not represent a valid integer".format(repr(string)))

def decode(string):
  value = unsafe_decode(string)
  validate_decoded(value)
  return value

def validate_decoded(integer):
  if integer < 0:
    raise gfapy.ValueError(
      "{} is not a positive integer".format(integer))

def validate_encoded(string):
  if not re.match(r"^[0-9]+$", string):
    raise gfapy.FormatError(
      "{} does not represent a valid unsigned integer".format(repr(string)))

def unsafe_encode(obj):
  return str(obj)

def encode(obj):
  if isinstance(obj, str):
    validate_encoded(obj)
    return obj
  elif isinstance(obj, int):
    validate_decoded(obj)
    return str(obj)
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, int)")



import gfapy

class Coverage:

  def coverage(self, count_tag = "RC", unit_length = 1):
    """Compute the coverage from the value a count_tag (RC, KC or FC).

    If unit_length is provided then: count/(length-unit_length+1),
    otherwise: count/length. The latter is a good approximation if
    length >>> unit_length.

    Parameters:
      count_tag (str): integer tag from which the count shall be
        taken (defaults to RC)
      unit_length (int): average length of the sequence which is counted
        (read for RC, fragment for FC, k-mer for KC).

    Returns:
      int : Coverage, if count_tag and length are defined.
      None : Otherwise
    """
    if count_tag in self.tagnames and self.length:
      return (float(self.get(count_tag)))/(self.length - unit_length + 1)
    else:
      return None

  def try_get_coverage(self, count_tag = "RC", unit_length = 1):
    """
    As coverage, but raises an exception if the coverage cannot be computed.
    """
    c = self.coverage(count_tag = count_tag, unit_length = unit_length)
    if c is None:
      self.try_get_length()
      raise gfapy.NotFoundError(
        "Tag {} undefined for segment {}".format(count_tag, self.name))
    else:
      return c



import gfapy
import unittest

class TestAPIRenameLines(unittest.TestCase):

  def test_rename(self):
    gfa = gfapy.Gfa(["S\t0\t*", "S\t1\t*", "S\t2\t*",
      "L\t0\t+\t2\t-\t12M", "C\t1\t+\t0\t+\t12\t12M", "P\t4\t2+,0-\t12M"])
    gfa.segment("0").name = "X"
    with self.assertRaises(gfapy.NotFoundError): gfa.try_get_segment("0")
    self.assertEqual(set(["X", "1", "2"]), set(gfa.segment_names))
    self.assertEqual("L\tX\t+\t2\t-\t12M", str(gfa.dovetails[0]))
    self.assertEqual("C\t1\t+\tX\t+\t12\t12M", str(gfa.containments[0]))
    self.assertEqual("P\t4\t2+,X-\t12M", str(gfa.paths[0]))
    with self.assertRaises(gfapy.NotFoundError): gfa.try_get_segment("0").dovetails_of_end("R")
    self.assertEqual("L\tX\t+\t2\t-\t12M", str(gfa.segment("X").dovetails_of_end("R")[0]))
    self.assertEqual("C\t1\t+\tX\t+\t12\t12M",
                 str(gfa.try_get_segment("1").edges_to_contained[0]))
    with self.assertRaises(gfapy.NotFoundError): gfa.try_get_segment("0").containers
    self.assertEqual("C\t1\t+\tX\t+\t12\t12M",
                 str(gfa.try_get_segment("X").edges_to_containers[0]))
    self.assertEqual("P\t4\t2+,X-\t12M", str(gfa.try_get_segment("X").paths[0]))




from .cloning import Cloning
from .connection import Connection
from .disconnection import Disconnection
from .dynamic_fields import DynamicFields
from .equivalence import Equivalence
from .field_data import FieldData
from .field_datatype import FieldDatatype
from .construction import Construction
from .update_references import UpdateReferences
from .validate import Validate
from .version_conversion import VersionConversion
from .virtual_to_real import VirtualToReal
from .writer import Writer



#!/usr/bin/env python3
"""
Prepare the output of the convert benchmark script for the R plotting script.
"""

import argparse
import os
import sys
import re

op = argparse.ArgumentParser(description=__doc__)
op.add_argument('--version', action='version', version='%(prog)s 1.0')
op.add_argument("--mult", "-m", action="store_true",
    help="set if variable n of edges/segment")
op.add_argument("inputfile")
opts = op.parse_args()

if not os.path.exists(opts.inputfile):
  sys.stderr.write("Input file not found: {}\n".format(opts.inputfile))
  exit(1)

with open(opts.inputfile) as inputfile:
  header = True
  if opts.mult:
    outdata = ["mult", "time", "space", "time_per_line", "space_per_line"]
  else:
    outdata = ["lines", "time", "space", "time_per_line", "space_per_line"]
  print("\t".join(outdata))
  for line in inputfile:
    if line[:3] == "###":
      header = False
    elif not header:
      data = line.rstrip("\n\r").split("\t")
      n_segments = data[2]
      multiplier = data[3]
      n_lines = int(int(n_segments) * (1+float(multiplier)))
      elapsed = data[5]
      elapsed_match = re.compile(r'\s+(\d+):(\d+\.\d+)').match(elapsed)
      if elapsed_match:
        minutes = int(elapsed_match.groups()[0])
        seconds = float(elapsed_match.groups()[1])
        seconds += minutes * 60
      else:
        elapsed_match = re.compile(r'\s+(\d+):(\d+):(\d+)').match(elapsed)
        if elapsed_match:
          hours = int(elapsed_match.groups()[0])
          minutes = int(elapsed_match.groups()[1])
          seconds = int(elapsed_match.groups()[2])
          minutes += hours * 60
          seconds += minutes * 60
        else:
          continue
      memory = data[6]
      memory = int(re.compile(r'(\d+) kB').match(memory).groups()[0])
      megabytes = memory / 1024
      if opts.mult:
        outdata = [str(multiplier)]
      else:
        outdata = [str(n_lines)]
      outdata += [str(seconds),str(megabytes),
                  str(seconds/n_lines), str(megabytes/n_lines)]
      print("\t".join(outdata))



import gfapy
import unittest

class TestGraphOpArtifacts(unittest.TestCase):

  def test_remove_small_components(self):
    for sfx in ["gfa", "gfa2"]:
      g = gfapy.Gfa.from_file("tests/testdata/two_components.{}".format(sfx))
      self.assertEqual(2, len(g.connected_components()))
      g.remove_small_components(1000)
      self.assertEqual(2, len(g.connected_components()))
      g.remove_small_components(3000)
      self.assertEqual(1, len(g.connected_components()))
      g.remove_small_components(10000)
      self.assertEqual(0, len(g.connected_components()))

  def test_remove_dead_ends(self):
    for sfx in ["gfa", "gfa2"]:
      g = gfapy.Gfa.from_file("tests/testdata/dead_ends.{}".format(sfx))
      self.assertEqual(6, len(g.segments))
      g.remove_dead_ends(100)
      self.assertEqual(6, len(g.segments))
      g.remove_dead_ends(1500)
      self.assertEqual(5, len(g.segments))
      g.remove_dead_ends(1500)
      self.assertEqual(5, len(g.segments))
      g.remove_dead_ends(150000)
      g.remove_dead_ends(150000)
      self.assertEqual(2, len(g.segments))
      g.remove_dead_ends(1500000)
      self.assertEqual(0, len(g.segments))




import gfapy
import unittest

class TestgfapyGfaToolsLinearPaths(unittest.TestCase):

  def test_linear_path_merging(self):
    for sfx in ["gfa", "gfa2"]:
      gfa = gfapy.Gfa.from_file("tests/testdata/linear_merging.2."+"{}".format(sfx))
      gfa.merge_linear_path([["0", "R"],["1", "R"],["2", "L"],["3", "R"]],
                            enable_tracking=True)
      gfa.try_get_segment("0_1_2^_3") # nothing raised
      self.assertEqual("ACGACGACGTCGA", gfa.segment("0_1_2^_3").sequence)
      gfa = gfapy.Gfa.from_file("tests/testdata/linear_merging.2."+"{}".format(sfx))
      gfa.merge_linear_path([["0", "R"],["1", "R"],["2", "L"],["3", "R"]],
                            enable_tracking=True)
      gfa.try_get_segment("0_1_2^_3") # nothing raised
      self.assertEqual("ACGACGACGTCGA", gfa.segment("0_1_2^_3").sequence)

  def test_linear_path_merge_all(self):
    for sfx in ["gfa", "gfa2"]:
      gfa = gfapy.Gfa.from_file("tests/testdata/linear_merging.3.{}".format(sfx))
      gfa.merge_linear_paths(enable_tracking=True)
      self.assertIn(gfa.segment_names[0], ["0_1_2^_3","3^_2_1^_0^"])
      gfa = gfapy.Gfa.from_file("tests/testdata/linear_merging.4.{}".format(sfx))
      gfa.merge_linear_paths(enable_tracking=True)
      try:
        self.assertEqual({"0","3","1_2^"}, {x.name for x in gfa.segments})
      except:
        self.assertEqual({"0","3","2_1^"}, {x.name for x in gfa.segments})




import gfapy
import unittest

class TestApiCustomRecords(unittest.TestCase):

  def test_from_string(self):
    str1 = "X\tthis is a\tcustom line"
    l1 = gfapy.Line(str1)
    self.assertEqual(gfapy.line.CustomRecord, l1.__class__)
    self.assertEqual("X", l1.record_type)
    self.assertEqual("this is a", l1.field1)
    self.assertEqual("custom line", l1.field2)

  def test_from_string_with_tags(self):
    str2 = "XX\txx:i:2\txxxxxx\txx:i:1"
    l2 = gfapy.Line(str2)
    self.assertEqual(gfapy.line.CustomRecord, l2.__class__)
    self.assertEqual("XX", l2.record_type)
    self.assertEqual("xx:i:2", l2.field1)
    self.assertEqual("xxxxxx", l2.field2)
    with self.assertRaises(AttributeError): l2.field3
    self.assertEqual(1, l2.xx)
    l2.xx = 3
    self.assertEqual(3, l2.xx)
    l2.field1 = "blabla"
    self.assertEqual("blabla", l2.field1)

  def test_to_s(self):
    str1 = "X\tthis is a\tcustom line"
    self.assertEqual(str1, str(gfapy.Line(str1)))
    str2 = "XX\txx:i:2\txxxxxx\txx:i:1"
    self.assertEqual(str2, str(gfapy.Line(str2)))

  def test_add_custom_records(self):
    gfa = gfapy.Gfa(version="gfa2")
    x1 = "X\tthis is a custom record"
    gfa.append(x1) # nothing raised
    self.assertEqual(["X"], gfa.custom_record_keys)
    self.assertEqual([x1], [str(x) for x in gfa.custom_records_of_type("X")])

  def test_delete_custom_records(self):
    gfa = gfapy.Gfa(version="gfa2")
    c = "X\tThis is a custom_record"
    gfa.append(c)
    self.assertEqual([c], [str(x) for x in gfa.custom_records_of_type("X")])
    for x in gfa.custom_records_of_type("X"): x.disconnect()
    self.assertEqual([], gfa.custom_records_of_type("X"))

  def test_custom_records(self):
    x = ["X\tVN:Z:1.0", "Y\ttesttesttest"]
    self.assertEqual(x[0], str(gfapy.Gfa(x).custom_records_of_type("X")[0]))
    self.assertEqual(x[1], str(gfapy.Gfa(x).custom_records_of_type("Y")[0]))




import unittest
import gfapy

class TestUnitLine(unittest.TestCase):

  def test_initialize_not_enough_positional_fields(self):
    gfapy.Line(["S", "1", "*"])
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line(["S", "1"])

  def test_initialize_too_many_positionals(self):
    with self.assertRaises(gfapy.FormatError):
      gfapy.Line(["S", "1", "*", "*"])

  def test_initialize_predefined_tag_wrong_type(self):
    gfapy.line.Header(["H", "VN:Z:1"])
    with self.assertRaises(gfapy.TypeError):
      gfapy.line.Header(["H", "VN:i:1"])

  def test_initialize_wrong_tag_format(self):
    with self.assertRaises(gfapy.FormatError):
      gfapy.line.Header(["H", "VN i:1"])

  def test_initialize_positional_field_type_error(self):
    with self.assertRaises(gfapy.FormatError):
      gfapy.line.segment.GFA1(["S", "1\t1", "*", "*"])

  def test_initialize_tag_type_error(self):
    with self.assertRaises(gfapy.FormatError):
      gfapy.line.Header(["H", "zz:i:1A"])

  def test_initialize_duplicate_tag(self):
    with self.assertRaises(gfapy.NotUniqueError):
      gfapy.line.Header(["H", "zz:i:1", "zz:i:2"])
    with self.assertRaises(gfapy.NotUniqueError):
      gfapy.line.Header(["H", "zz:i:1", "VN:Z:1", "zz:i:2"])

  def test_initialize_custom_tag(self):
    gfapy.line.Header(["H", "ZZ:Z:1"]) # nothing raised

  def test_record_type(self):
    l = gfapy.line.Header(["H", "xx:i:13", "VN:Z:HI"])
    self.assertEqual("H", l.record_type)
    with self.assertRaises(AttributeError):
      l.record_type = "S"

  def test_add_tag(self):
    l = gfapy.line.Header(["H", "xx:i:13", "VN:Z:HI"])
    self.assertEqual(None, l.xy)
    l.set("xy", "HI")
    self.assertEqual("HI", l.xy)

  def test_unknown_record_type(self):
    with self.assertRaises(gfapy.VersionError):
      gfapy.Line("Z\txxx", version = "gfa1")
    gfapy.Line("Z\txxx", version = "gfa2")
    gfapy.Line("Z\txxx")

  def test_field_alias(self):
    s = gfapy.Line("S\tA\t*")
    self.assertEqual(s.name, s.get("name"))
    self.assertEqual("A", s.name)
    self.assertEqual("A", s.sid)
    self.assertEqual("A", s.get("name"))
    self.assertEqual("A", s.get("sid"))
    s.set("name", "B")
    self.assertEqual("B", s.get("sid"))
    s.set("sid", "C")
    self.assertEqual("C", s.name)

  def test_to_s(self):
    fields = ["H", "VN:Z:HI", "xx:i:13"]
    l = gfapy.line.Header(fields[:])
    lstr = str(l)
    self.assertEqual("\t".join(fields), lstr)

  def test_clone(self):
    l = gfapy.Line("H\tVN:Z:1.0")
    l1 = l
    l2 = l.clone()
    self.assertIsInstance(l, gfapy.line.Header)
    self.assertIsInstance(l2, gfapy.line.Header)
    l2.VN = "2.0"
    self.assertEqual("2.0", l2.VN)
    self.assertEqual("1.0", l.VN)
    l1.VN = "2.0"
    self.assertEqual("2.0", l.VN)




import unittest
import gfapy

class TestUnitHeader(unittest.TestCase):

  def test_new(self):
    gfapy.line.Header(["H", "VN:Z:1.0", "xx:i:11"])

  def test_string_to_gfa_line(self):
    gfapy.Line("H\tVN:Z:1.0")
    assert(isinstance(gfapy.Line("H\tVN:Z:1.0"),gfapy.line.Header))
    self.assertEqual(gfapy.line.Header(["H", "VN:Z:1.0", "xx:i:11"]),
      gfapy.Line("H\tVN:Z:1.0\txx:i:11"))
    self.assertRaises(gfapy.FormatError,
      gfapy.Line, "H\tH2\tVN:Z:1.0")
    self.assertRaises(gfapy.TypeError,
      gfapy.Line, "H\tVN:i:1.0")

  def test_to_s(self):
    try:
      self.assertEqual("H\tVN:Z:1.0\txx:i:11",
        str(gfapy.line.Header(["H", "VN:Z:1.0", "xx:i:11"])))
    except:
      self.assertEqual("H\txx:i:11\tVN:Z:1.0",
        str(gfapy.line.Header(["H", "VN:Z:1.0", "xx:i:11"])))

  def test_tag_reading(self):
    self.assertEqual("1.0",
      gfapy.line.Header(["H", "VN:Z:1.0", "xx:i:11"]).VN)

  def test_tag_writing(self):
    gfapy.line.Header(["H", "VN:Z:1.0", "xx:i:11"]).VN = "2.0"

  def test_connection(self):
    assert(not gfapy.line.Header(["H"]).is_connected())
    assert(gfapy.Gfa().header.is_connected())
    self.assertRaises(gfapy.RuntimeError,
      gfapy.line.Header(["H"]).connect, gfapy.Gfa())

  def test_to_gfa1_a(self):
    line = gfapy.Line("H\tVN:Z:1.0\txx:i:1")
    self.assertEqual("H", line._to_gfa1_a()[0])
    self.assertEqual(sorted(["VN:Z:1.0", "xx:i:1"]), sorted(line._to_gfa1_a()[1:]))
    line = gfapy.Line("H\tVN:Z:2.0\txx:i:1")
    self.assertEqual("H", line._to_gfa1_a()[0])
    self.assertEqual(sorted(["VN:Z:1.0", "xx:i:1"]), sorted(line._to_gfa1_a()[1:]))

  def test_to_gfa2_a(self):
    line = gfapy.Line("H\tVN:Z:1.0\txx:i:1")
    self.assertEqual("H", line._to_gfa2_a()[0])
    self.assertEqual(sorted(["VN:Z:2.0", "xx:i:1"]), sorted(line._to_gfa2_a()[1:]))
    line = gfapy.Line("H\tVN:Z:2.0\txx:i:1")
    self.assertEqual("H", line._to_gfa2_a()[0])
    self.assertEqual(sorted(["VN:Z:2.0", "xx:i:1"]), sorted(line._to_gfa2_a()[1:]))

  def test_add(self):
    line = gfapy.Line("H\tVN:Z:2.0\txx:i:1")
    line.add("yy", "test")
    self.assertEqual("test", line.yy)
    line.add("yy", "test")
    self.assertEqual(["test","test"], line.yy)
    line.add("yy", "test")
    self.assertEqual(["test","test","test"], line.yy)
    line.add("VN", "2.0")
    self.assertEqual("2.0", line.VN)
    self.assertRaises(gfapy.InconsistencyError, line.add, "VN", "1.0")
    line.add("TS", "120")
    self.assertEqual(120, line.TS)
    line.add("TS", 120)
    line.add("TS", "120")
    self.assertRaises(gfapy.InconsistencyError, line.add, "TS", 130)
    self.assertRaises(gfapy.InconsistencyError, line.add, "TS", "140")

  def test_field_to_s(self):
    line = gfapy.Line("H\tVN:Z:1.0\txx:i:1")
    line.add("xx", 2)
    self.assertEqual("1.0", line.field_to_s("VN"))
    self.assertEqual("1\t2", line.field_to_s("xx"))
    self.assertEqual("VN:Z:1.0", line.field_to_s("VN", tag=True))
    self.assertEqual("xx:i:1\txx:i:2", line.field_to_s("xx", tag=True))

  def test_n_duptags(self):
    line = gfapy.Line("H\tVN:Z:1.0\txx:i:1")
    self.assertEqual(0, line._n_duptags())
    line.add("xx", 2)
    self.assertEqual(1, line._n_duptags())
    line.add("xx", 2)
    self.assertEqual(1, line._n_duptags())
    line.add("zz", 2)
    self.assertEqual(1, line._n_duptags())
    line.add("zz", 2)
    self.assertEqual(2, line._n_duptags())

  def test_split(self):
    line = gfapy.Line("H\tVN:Z:1.0\txx:i:1")
    line.add("xx", 2)
    self.assertEqual(3, len(line._split()))
    for s in line._split():
      assert(isinstance(s, gfapy.line.Header))
    self.assertEqual(sorted(["H\tVN:Z:1.0", "H\txx:i:1", "H\txx:i:2"]),
                     sorted([str(x) for x in line._split()]))

  def test_merge(self):
    line1 = gfapy.Line("H\tVN:Z:1.0\txx:i:1")
    line2 = gfapy.Line("H\txx:i:2\tyy:f:1.0")
    line1._merge(line2)
    self.assertEqual("1.0", line1.VN)
    self.assertEqual([1,2], line1.xx)
    self.assertEqual(1.0, line1.yy)




import gfapy
import re

def unsafe_decode(string):
  return [ gfapy.OrientedLine(str(l[0:-1]), str(l[-1]))
           for l in string.split(" ")]

def decode(string):
  validate_encoded(string)
  return unsafe_decode(string)

def validate_encoded(string):
  if not re.match(r"^[!-~]+[+-]( [!-~]+[+-])*$", string):
    raise gfapy.FormatError(
      "{} is not a valid list of GFA2 segment names ".format(repr(string))+
      "and orientations")

def validate_decoded(iterable):
  for elem in iterable:
    if not isinstance(elem, gfapy.OrientedLine):
      raise gfapy.TypeError(
            "the list contains an object of class {}\n".format(type(elem))+
            "(accepted classes: gfapy.OrientedLine)")
    elem.validate()
    if not re.match(r"^[!-~]+$", elem.name):
      raise gfapy.FormatError(
        "the list contains an invalid GFA2 identifier {}\n".format(elem.name)+
        "(it contains spaces and/or non-printable characters)")
    if not elem.orient in ["+", "-"]:
      raise gfapy.FormatError(
        "{} is not a valid orientation".format(elem.orient))

def unsafe_encode(obj):
  if isinstance(obj, str):
    return obj
  elif isinstance(obj, list):
    retval = []
    for elem in obj:
      if not isinstance(elem, gfapy.OrientedLine):
        raise gfapy.TypeError(
              "the list contains an object of class {}\n".format(type(elem))+
              "(accepted classes: gfapy.OrientedLine)")
      retval.append(str(elem))
    return " ".join(retval)
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: str, list)")

def encode(obj):
  validate_decoded(obj)
  return unsafe_encode(obj)



import gfapy

class GFA2ToGFA1:

  def _to_gfa1_a(self, slen_tag = "LN"):
    """
    Notes:
      According to the GFA2 specification, slen must be not be really the length
      of the sequence. By default, this is ignored, and the content of slen is
      stored in the LN tag.

    Parameters:
      slen_tag (str) : tag to use in GFA1 to store the content of slen

    Returns:
      str list : A array of GFA1 field strings.
    """
    a = ["S", self.field_to_s("name", tag = False),
              self.field_to_s("sequence", tag = False)]
    a.append(gfapy.Field._to_gfa_tag(self.slen, slen_tag, datatype = "i"))
    for fn in self.tagnames:
      a.append(self.field_to_s(fn, tag = True))
    return a



import gfapy
from functools import total_ordering

@total_ordering
class LastPos:
  """The last position of a sequence.

  In GFA2 lines (e.g. edges and fragments), the last position of a sequence is
  suffixed by a ``$`` sign.

  This class is provided to represent this particular case of a position value.
  New instances are created passing an integer or the string representation
  to the constructor. If a string without ``$`` is passed to the constructor,
  an integer is returned instead (thus the constructor can be used to parse
  any GFA position field string).

  >>> from gfapy import LastPos
  >>> LastPos("2$")
  gfapy.LastPos(2)
  >>> LastPos(2)
  gfapy.LastPos(2)
  >>> LastPos("2")
  2

  Parameters:
    value (str, int) : a string representation of a position, or an integer
      representing the last position of a sequence
    valid (bool) : if True, the value is always considered valid, and no
      validation is performed (default: False)

  Returns:
    gfapy.LastPos, int : an int if the value is a string and has no dollar sign
      suffix; otherwise a LastPos instance

  Raises:
    see validate method (no exceptions raised if valid is set to True)

  """

  def __new__(cls, value, valid=False):
    if isinstance(value, str):
      return cls._from_string(value, valid=valid)
    else:
      new_instance = object.__new__(cls)
      new_instance.value = value
      if not valid:
        new_instance.validate()
      return new_instance

  def validate(self):
    """Checks that the value is a positive integer.

    Validation is performed by default on construction, unless the valid
    parameter of the constructor is set to True.

    Raises:
      gfapy.error.TypeError : if the value is not an integer
      gfapy.error.ValueError : if the value is not >= 0
    """

    if not isinstance(self.value, int):
      raise gfapy.TypeError("LastPos value shall be an integer,"+
          " {} found".format(type(self.value)))
    elif self.value < 0:
      raise gfapy.ValueError("LastPos value shall be >= 0,"+
          " {} found".format(self.value))

  def __str__(self):
    return "{}$".format(self.value)

  def __repr__(self):
    return "gfapy.LastPos({})".format(self.value)

  def __int__(self):
    return self.value

  def __eq__(self, other):
    if isinstance(other, int):
      return self.value == other
    elif not isinstance(other, LastPos):
      return NotImplemented
    else:
      return self.value == other.value

  def __lt__(self, other):
    if isinstance(other, int):
      return self.value.__lt__(other)
    elif not isinstance(other, LastPos):
      return NotImplemented
    else:
      return self.value.__lt__(other.value)

  def __sub__(self,other):
    o = int(other)
    if o == 0:
      return gfapy.LastPos(self.value)
    else:
      return self.value - o

  @classmethod
  def _from_string(cls, string, valid=False):
    if string[-1] == "$":
      return cls(int(string[:-1]), valid=valid)
    else:
      try:
        v = int(string)
      except:
        raise gfapy.FormatError(
            "LastPos value has a wrong format: {}".format(string))
      if not valid:
        if v < 0:
          raise gfapy.ValueError("LastPos value shall be >= 0,"+
              " {} found".format(v))
      return v

def posvalue(obj):
  """The integer representing a position.

  Parameters:
    obj (int, LastPos) : the position

  Returns:
    int : If obj is a LastPos, then its value.
          If it is an integer, then the integer itself.
  """
  if isinstance(obj, LastPos):
    return obj.value
  else:
    return obj

def islastpos(obj):
  """Checks if a position value is a last position.

  Parameters:
    obj (int, LastPos) : the position

  Returns:
    bool : If obj is a LastPos, then True.
           If it is an integer, then False.
  """
  return isinstance(obj, LastPos)

def isfirstpos(obj):
  """Checks if a position value is the first position (0).

  Note that the last position of an empty sequence
  is also its first position, therefore:

  >>> from gfapy.lastpos import isfirstpos
  >>> isfirstpos(gfapy.LastPos("0$"))
  True

  Parameters:
    obj (int, LastPos) : the position

  Returns:
    bool : If the value of the position is 0, then True.
  """
  return posvalue(obj) == 0




from .gfa1_to_gfa2 import GFA1ToGFA2
from .length_gfa1 import LengthGFA1
from .coverage import Coverage
from .references import References
from .writer_wo_sequence import WriterWoSequence
from . import Segment

class GFA1(WriterWoSequence, References, Coverage,
           LengthGFA1, GFA1ToGFA2, Segment):
  """
  A segment line of a GFA file
  """

  VERSION = "gfa1"
  RECORD_TYPE = "S"
  POSFIELDS = ["name", "sequence"]
  PREDEFINED_TAGS = ["LN", "RC", "FC", "KC", "SH", "UR"]
  DATATYPE = {
    "name" : "segment_name_gfa1",
    "sequence" : "sequence_gfa1",
    "LN" : "i",
    "RC" : "i",
    "FC" : "i",
    "KC" : "i",
    "SH" : "H",
    "UR" : "Z",
  }
  NAME_FIELD = "name"
  FIELD_ALIAS = { "sid" : "name" }
  DEPENDENT_LINES = ["dovetails_L", "dovetails_R",
                     "edges_to_contained", "edges_to_containers", "paths"]
  gfa2_compatibility = ["gaps_L", "gaps_R", "fragments", "internals", "sets"]
  OTHER_REFERENCES = gfa2_compatibility

GFA1._apply_definitions()



import unittest
import gfapy

class TestUnitSegmentEnd(unittest.TestCase):

  sym = "a"
  ref = gfapy.Line("S\ta\t*\txx:Z:1.0")
  invalid_sym = "a\ta"
  invalid_ref = []
  se_s = gfapy.SegmentEnd(sym, "L")
  se_r = gfapy.SegmentEnd(ref, "R")
  se_s_str = "aL"
  se_r_str = "aR"
  se_s_sym = "aL"
  se_r_sym = "aR"

  def test_new(self):
    gfapy.SegmentEnd(TestUnitSegmentEnd.sym, "L")
    # no validation on creation
    gfapy.SegmentEnd(TestUnitSegmentEnd.invalid_sym, "X")

  def test_from_list(self):
    self.assertEqual(TestUnitSegmentEnd.se_s,
        gfapy.SegmentEnd(["a", "L"]))
    self.assertEqual(gfapy.SegmentEnd,
        gfapy.SegmentEnd(["a", "L"]).__class__)
    self.assertRaises(gfapy.ArgumentError, gfapy.SegmentEnd,
      ["a", "L", "L"])
    gfapy.SegmentEnd(["a", "X"]) # no validation

  def test_segment(self):
    self.assertEqual(TestUnitSegmentEnd.sym, TestUnitSegmentEnd.se_s.segment)
    self.assertEqual(TestUnitSegmentEnd.ref, TestUnitSegmentEnd.se_r.segment)
    se2 = gfapy.SegmentEnd(TestUnitSegmentEnd.sym, "R")
    se2.segment = TestUnitSegmentEnd.ref
    self.assertEqual(TestUnitSegmentEnd.ref, se2.segment)

  def test_end_type(self):
    self.assertEqual("L", TestUnitSegmentEnd.se_s.end_type)
    self.assertEqual("R", TestUnitSegmentEnd.se_r.end_type)
    se2 = gfapy.SegmentEnd(TestUnitSegmentEnd.sym, "L")
    se2.end_type = "R"
    self.assertEqual("R", se2.end_type)

  def test_name(self):
    self.assertEqual(TestUnitSegmentEnd.sym, TestUnitSegmentEnd.se_s.name)
    self.assertEqual(TestUnitSegmentEnd.sym, TestUnitSegmentEnd.se_r.name)

  def test_validate(self):
    TestUnitSegmentEnd.se_s.validate()
    TestUnitSegmentEnd.se_r.validate()
    se1 = gfapy.SegmentEnd("a", "X")
    self.assertRaises(gfapy.ValueError, se1.validate)

  def test_inverted(self):
    inv_s = TestUnitSegmentEnd.se_s.inverted()
    self.assertEqual(TestUnitSegmentEnd.se_s.segment, inv_s.segment)
    self.assertEqual("R", inv_s.end_type)
    inv_r = TestUnitSegmentEnd.se_r.inverted()
    self.assertEqual(TestUnitSegmentEnd.se_r.segment, inv_r.segment)
    self.assertEqual("L", inv_r.end_type)

  def test_to_s(self):
    self.assertEqual(TestUnitSegmentEnd.se_s_str, str(TestUnitSegmentEnd.se_s))
    self.assertEqual(TestUnitSegmentEnd.se_r_str, str(TestUnitSegmentEnd.se_r))

  def test_equal(self):
    se2 = gfapy.SegmentEnd(TestUnitSegmentEnd.sym, "L")
    se3 = gfapy.SegmentEnd(TestUnitSegmentEnd.ref, "R")
    self.assertEqual(TestUnitSegmentEnd.se_s, se2)
    self.assertEqual(TestUnitSegmentEnd.se_r, se3)
    # only name and end_type equivalence is checked, not segment
    assert(TestUnitSegmentEnd.se_r != TestUnitSegmentEnd.se_s)
    assert(TestUnitSegmentEnd.se_r.inverted() == TestUnitSegmentEnd.se_s)
    # equivalence to array
    assert(TestUnitSegmentEnd.se_s == ["a","L"])
    assert(TestUnitSegmentEnd.se_r == ["a","R"])

  #def test_comparison(self):
  #  self.assertEqual(-1, ["a","L"].to_segment_end() <=> ["b","L"].to_segment_end())
  #  self.assertEqual(0,  ["a","L"].to_segment_end() <=> ["a","L"].to_segment_end())
  #  self.assertEqual(1,  ["b","L"].to_segment_end() <=> ["a","L"].to_segment_end())
  #  self.assertEqual(-1, ["a","L"].to_segment_end() <=> ["a","R"].to_segment_end())
  #  self.assertEqual(0,  ["a","R"].to_segment_end() <=> ["a","R"].to_segment_end())
  #  self.assertEqual(1,  ["a","R"].to_segment_end() <=> ["a","L"].to_segment_end())

  def test_segment_ends_path(self):
    sep = gfapy.SegmentEndsPath([gfapy.SegmentEnd("a","L"),
                                 gfapy.SegmentEnd("b","R")])
    self.assertEqual([gfapy.SegmentEnd("b","L"),gfapy.SegmentEnd("a","R")],
      list(reversed(sep)))
    self.assertNotEqual([gfapy.SegmentEnd("b","L"),gfapy.SegmentEnd("a","R")],
      sep)
    sep.reverse()
    self.assertEqual([gfapy.SegmentEnd("b","L"),gfapy.SegmentEnd("a","R")],
      sep)



import unittest
import gfapy

class TestLineEdge(unittest.TestCase):

  def test_to_gfa1(self):
    e1 = gfapy.Line("E\t*\t1+\t2+\t90\t100$\t0\t10\t10M")
    l1 = "L\t1\t+\t2\t+\t10M"
    self.assertEqual(l1, e1.to_gfa1_s())
    e2 = gfapy.Line("E\t*\t1+\t2+\t0\t20\t80\t100$\t20M")
    l2 = "L\t2\t+\t1\t+\t20M"
    self.assertEqual(l2, e2.to_gfa1_s())
    e3 = gfapy.Line("E\t*\t3-\t4+\t0\t30\t0\t30\t30M")
    l3 = "L\t3\t-\t4\t+\t30M"
    self.assertEqual(l3, e3.to_gfa1_s())
    e4 = gfapy.Line("E\t*\t3+\t4-\t60\t100$\t60\t100$\t40M")
    l4 = "L\t3\t+\t4\t-\t40M"
    self.assertEqual(l4, e4.to_gfa1_s())



from ..line import Line
import gfapy
import re

class Segment(Line):
  """
  Parent class for classes representing segment lines
  """

  @staticmethod
  def _subclass(data):
    n_positionals = len(data)-1
    for i in range(len(data)-1, 0, -1):
      if not re.search(r"^..:.:.*$", data[i]):
        break
      n_positionals = i-1
    if n_positionals == 2:
      return gfapy.line.segment.GFA1
    elif n_positionals == 3:
      return gfapy.line.segment.GFA2
    else:
      raise gfapy.FormatError("Wrong number of positional fields for "
          "segment line; GFA1=2, GFA2=3, found={}\n".format(n_positionals))



import gfapy

def unsafe_decode(string):
  return gfapy.Alignment(string, valid = True, version = "gfa2")

def decode(string):
  return gfapy.Alignment(string, valid = False, version = "gfa2")

validate_encoded = decode

def validate_decoded(alignment):
  alignment.validate()

def unsafe_encode(obj):
  return str(obj)

def encode(obj):
  if isinstance(obj, str):
    validate_encoded(obj)
    return obj
  elif isinstance(obj, gfapy.CIGAR) or isinstance(obj, gfapy.Trace):
    obj.validate()
    return str(obj)
  elif isinstance(obj, gfapy.Placeholder):
    return "*"
  else:
    raise gfapy.TypeError(
      "the class {} is incompatible with the datatype\n"
      .format(obj.__class__.__name__)+
      "(accepted classes: "+
      "str, CIGAR, Trace, AlignmentPlaceholder)")




import gfapy

class References:

  @property
  def dovetails(self):
    """
    References to the graph lines which involve the segment as dovetail overlap.

    Returns
    -------
    gfapy.line.Edge list
      A list of lines.
      The lines themselves can be modified, but the list is immutable.
    """
    return self.dovetails_L + self.dovetails_R

  def dovetails_of_end(self, extremity):
    """
    References to the graph lines which involve the segment as dovetail overlap.

    Returns
    -------
    gfapy.line.Edge list
      A list of lines.
      The lines themselves can be modified, but the list is immutable.
    """
    return getattr(self, "dovetails_{}".format(extremity))

  @property
  def gaps(self):
    """
    References to the gap lines which involve the segment.
    """
    return self.gaps_L + self.gaps_R

  def gaps_of_end(self, extremity):
    """
    References to the gap lines which involve the segment.
    """
    return getattr(self, "gaps_{}".format(extremity))

  @property
  def containments(self):
    """
    References to graph edges (C lines for GFA1, E for GFA2) which involve the
    segment in a containment relationship.
    """
    return self.edges_to_contained + self.edges_to_containers

  def _connectivity(self):
    """
    Computes the connectivity of a segment from its number of dovetail overlaps.

    Returns
    -------
    (conn_symbol,conn_symbol) list

    conn. symbols respectively of the :L and :R ends of +segment+.

    <b>Connectivity symbol:</b> (+conn_symbol+)
    - Let _n_ be the number of links to an end (+:L+ or +:R+) of a segment.
    Then the connectivity symbol is +:M+ if <i>n > 1</i>, otherwise _n_.
    """
    if not self.is_connected():
      raise gfapy.ArgumentError(
        "Cannot compute the connectivity of {}\n".format(self)+
        "Segment is not connected to a GFA instance")
    return self._connectivity_symbols(len(self.dovetails_L),
                                      len(self.dovetails_R))

  @property
  def neighbours(self):
    """
    List of dovetail-neighbours of a segment.

    Returns
    -------
    gfapy.line.Segment list
    	Segments connected to the current segment by dovetail overlap
      relationships (L lines for GFA1, dovetail-representing E lines for GFA2)
    """
    seen = set()
    return [l.other(self) for l in self.dovetails \
              if id(l) not in seen and not seen.add(id(l))]

  @property
  def neighbours_L(self):
    """
    List of dovetail-neighbours of a segment.

    Returns
    -------
    gfapy.line.Segment list
    	Segments connected to the current segment by dovetail overlap
      relationships (L lines for GFA1, dovetail-representing E lines for GFA2)
    """
    seen = set()
    return [l.other(self) for l in self.dovetails_L \
             if id(l) not in seen and not seen.add(id(l))]

  @property
  def neighbours_R(self):
    """
    List of dovetail-neighbours of a segment.

    Returns
    -------
    gfapy.line.Segment list
    	Segments connected to the current segment by dovetail overlap
      relationships (L lines for GFA1, dovetail-representing E lines for GFA2)
    """
    seen = set()
    return [l.other(self) for l in self.dovetails_R \
             if id(l) not in seen and not seen.add(id(l))]

  def neighbours_of_end(self, extremity):
    return getattr(self, "neighbours_{}".format(extremity))

  @property
  def containers(self):
    """
    List of segments which contain the segment.

    Returns
    -------
    gfapy.line.Segment list
    	Segments connected to the current segment by containment relationships
      (C lines for GFA1, containment-representing E lines for GFA2),
      where the current segment is the contained segment.
    """
    seen = set()
    return [l.from_segment for l in self.edges_to_containers \
             if id(l) not in seen and not seen.add(id(l))]

  @property
  def contained(self):
    """
    List of segments which are contained in the segment.

    Returns
    -------
    gfapy.line.Segment list
    	Segments connected to the current segment by containment relationships
      (C lines for GFA1, containment-representing E lines for GFA2),
      where the current segment is the container segment.
    """
    seen = set()
    return [l.to_segment for l in self.edges_to_contained \
             if id(l) not in seen and not seen.add(id(l))]

  @property
  def edges(self):
    """
    List of edges which refer to the segment

    Returns
    -------
    gfapy.line.Edge list
    """
    return self.dovetails + self.containments + self.internals

  def relations_to(self, segment, collection="edges"):
    if isinstance(segment, gfapy.Line):
      return [e for e in getattr(self, collection) \
          if (e.other(self) is segment)]
    else:
      return [e for e in getattr(self, collection) \
          if (e.other(self).name == segment)]

  def oriented_relations(self, orientation, oriented_segment, collection="edges"):
    return [e for e in getattr(self, collection) if \
      (e.other_oriented_segment(gfapy.OrientedLine(self, orientation), tolerant=True) == \
        oriented_segment)]

  def end_relations(self, extremity, segment_end, collection ="edges"):
    return [e for e in getattr(self, collection) if \
      (e.other_end(gfapy.SegmentEnd(self, extremity), tolerant=True) == \
        segment_end)]

  def _connectivity_symbols(self, n, m):
    return (self._connectivity_symbol(n), self._connectivity_symbol(m))

  def _connectivity_symbol(self, n):
    return "M" if n > 1 else n

  def _backreference_keys(self, ref, key_in_ref):
    if ref.record_type == "E":
      return ["dovetails_L", "dovetails_R", "internals",
              "edges_to_containers", "edges_to_contained"]
    elif ref.record_type == "L":
      return ["dovetails_L", "dovetails_R"]
    elif ref.record_type == "C":
      return ["edges_to_contained"] if (key_in_ref == "from_segment") \
        else ["edges_to_containers"]
    elif ref.record_type == "G":
      return ["gaps_L", "gaps_R"]
    elif ref.record_type == "F":
      return ["fragments"]
    elif ref.record_type == "P" or ref.record_type == "O":
      return ["paths"]
    elif ref.record_type == "U":
      return ["sets"]
    else:
      return []



class CopyNumber:

  def set_default_count_tag(self, tag):
    """Set the count tag to be used by default for the coverage computation"""
    self._default["count_tag"] = tag

  def set_count_unit_length(self, unit_length):
    """Set the unit length to be used by default for the coverage computation"""
    self._default["unit_length"] = unit_length

  def delete_low_coverage_segments(self, mincov, count_tag=None,
                                   unit_length=None):
    """Remove the segments whose coverage is smaller than a specified value.

    Parameters:
      mincov (int) : the minimal coverage to keep a segment
      count_tag (str) : the name of the tag to use for coverage computation
      unit_length (int) : the unit length to use for coverage computation
    """
    if unit_length is None:
      unit_length = self._default["unit_length"]
    if count_tag is None:
      count_tag = self._default["count_tag"]
    for s in self.segments:
      cov = s.coverage(count_tag=count_tag, unit_length=unit_length)
      if cov < mincov:
        s.disconnect()

  def compute_copy_numbers(self, single_copy_coverage, mincov=None,
                           count_tag=None, cn_tag="cn", unit_length=None):
    """Compute the estimated copy numbers of all segments, from their coverage.

    Parameters:
      mincov (int) : the minimal coverage to assign copy number 1; if not
        specified, 1/4 of the single_copy_coverage is used
      single_copy_coverage : the coverage corresponding to a copy number of 1
      cn_tag (str) : the tag where to store the computed values (default: cn)
      count_tag (str) : the name of the tag to use for coverage computation
      unit_length (int) : the unit length to use for coverage computation
    """
    if mincov is None:
      mincov = single_copy_coverage * 0.25
    if count_tag is None:
      count_tag = self._default["count_tag"]
    if unit_length is None:
      unit_length = self._default["unit_length"]
    for s in self.segments:
      cov = s.try_get_coverage(count_tag=count_tag, unit_length=unit_length)
      if cov < mincov:
        cn = 0
      elif cov < single_copy_coverage:
        cn = 1
      else:
        cn = round(cov / single_copy_coverage)
      s.set(cn_tag, cn)

  def apply_copy_numbers(self, count_tag="cn", distribute="auto",
                         origin_tag="or", conserve_components=True):
    """Multiply each segment per its copy number.

    The copy number must be stored in a tag (default: cn). It can be computed
    e.g. using the compute_copy_numbers() method.

    Parameters:
      origin_tag (str) : the tag where to store the origin tracking
        (default: or); see multiply()
      distribute (str) : the value of the distribute parameter of multiply();
        see multiply()
      count_tag (str) : the name of the tag to use for coverage computation
      conserve_components (bool) : If True, segments with copy number 0 are
        not deleted, if their removal would split a connected component in two;
        thereby only dovetail overlaps are considered (default: False)
    """
    for s in sorted(self.segments, key=lambda s:s.try_get(count_tag)):
      self.multiply(s.name, s.get(count_tag), distribute=distribute,
               copy_names=None, conserve_components=conserve_components,
               origin_tag=origin_tag, track_origin=True)



import gfapy
import binascii

class ByteArray(bytes):
  """Array of unsigned byte values.

  The class is used for the representation of the data contained in H tags.
  The content of instances of the class is read-only. To edit the array,
  the instance must be cast to a list, edited, and casted back to ByteArray,
  as in the example below.

  Example:
    >>> import gfapy
    >>> a = gfapy.ByteArray([1,2,3])
    >>> a[0] = 0
    Traceback (most recent call last):
    ...
    TypeError: 'ByteArray' object does not support item assignment
    >>> a_lst = list(a)
    >>> a_lst[0] = 0
    >>> a = gfapy.ByteArray(a_lst)
    >>> str(a)
    '000203'

  Parameters:
    arg (string or bytes): If the argument is of type string,
        it has to be a valid hex string.

  Raises:
    gfapy.FormatError: If the argument is a string and has an invalid format.
    gfapy.ValueError: If the argument is not an string or a byte array.
  """

  def __new__(cls, arg):
    try:
      if isinstance(arg, str):
        if len(arg) == 0:
          raise gfapy.FormatError
        return bytes.__new__(cls, binascii.unhexlify(arg))
      else:
        return bytes.__new__(cls, arg)
    except binascii.Error:
      raise gfapy.FormatError
    except ValueError:
      raise gfapy.ValueError
    except TypeError:
      raise gfapy.ValueError

  def __str__(self):
    return str(binascii.hexlify(self), "utf8").upper()

  def validate(self):
    """Validates the content of the instance.

    The content is always valid, as values cannot be modified directly (see
    below) and trying to create from invalid data will raise an exception. So
    the validation method is only a placeholder which always does nothing.
    """
    pass

  def _default_gfa_tag_datatype(self):
    """GFA tag datatype to use by default"""
    return 'H'



from .references import References
from .validation import Validation
from ..line import Line

class Fragment(References, Validation, Line):
  """
  A fragment line of a GFA2 file
  """
  RECORD_TYPE = "F"
  POSFIELDS = ["sid", "external", "s_beg", "s_end", "f_beg", "f_end",
               "alignment"]
  PREDEFINED_TAGS = ["VN", "TS"]
  STORAGE_KEY = "external"
  DATATYPE = {
    "sid" : "identifier_gfa2",
    "external" : "oriented_identifier_gfa2",
    "s_beg" : "position_gfa2",
    "s_end" : "position_gfa2",
    "f_beg" : "position_gfa2",
    "f_end" : "position_gfa2",
    "alignment" : "alignment_gfa2",
    "TS" : "i",
  }
  REFERENCE_FIELDS = ["sid"]

Fragment._apply_definitions()



import gfapy

class Tags:

  def set(self, fieldname, value):
    """Set the value of a field.

    The generic Line.set() method is overwritten for comments,
    in order to disallow tags.
    """
    if fieldname in ["content", "spacer"]:
      return super().set(fieldname, value)
    else:
      raise gfapy.RuntimeError("Tags of comment lines cannot be set")



import gfapy
import unittest

class TestInternalsFieldValidator(unittest.TestCase):

  def test_field_gfa_field_validate_i(self):
    gfapy.Field._validate_gfa_field("1"  , "i")
    gfapy.Field._validate_gfa_field("12" , "i")
    gfapy.Field._validate_gfa_field("-12", "i")
    self.assertRaises(gfapy.FormatError, gfapy.Field._validate_gfa_field, "1A", "i")
    self.assertRaises(gfapy.FormatError, gfapy.Field._validate_gfa_field, "A1", "i")
    self.assertRaises(gfapy.FormatError, gfapy.Field._validate_gfa_field, "2.1", "i")

  def test_field_gfa_field_validate_A(self):
    gfapy.Field._validate_gfa_field("A", "A")
    self.assertRaises(gfapy.FormatError, gfapy.Field._validate_gfa_field, "AA", "A")

  def test_field_gfa_field_validate_f(self):
    gfapy.Field._validate_gfa_field("-12.1", "f")
    gfapy.Field._validate_gfa_field("-12.1E-2", "f")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._validate_gfa_field("2.1X", "f")

  def test_field_gfa_field_validate_Z(self):
    gfapy.Field._validate_gfa_field("-12.1E-2", "Z")

  def test_field_gfa_field_validate_H(self):
    gfapy.Field._validate_gfa_field("0A12121EFF", "H")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._validate_gfa_field("21X1", "H")

  def test_field_gfa_field_validate_B(self):
    gfapy.Field._validate_gfa_field("i,12,-5", "B")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._validate_gfa_field("C,X1", "B")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._validate_gfa_field("f.1.1", "B")

  def test_field_gfa_field_validate_J(self):
    gfapy.Field._validate_gfa_field("{\"1\":2}", "J")
    with self.assertRaises(gfapy.FormatError):
      gfapy.Field._validate_gfa_field("1\t2", "J")



import unittest
from copy import deepcopy
import gfapy

class TestApiAlignment(unittest.TestCase):

  cigar_1 = gfapy.CIGAR([
      gfapy.CIGAR.Operation(12,"M"),
      gfapy.CIGAR.Operation(1,"D"),
      gfapy.CIGAR.Operation(2,"I"),
      gfapy.CIGAR.Operation(0,"M"),
      gfapy.CIGAR.Operation(1,"P")])
  cigar_1_s = "12M1D2I0M1P"

  cigar_gfa1_1_s = "1S2M3I4=5X6D7P8N9H"
  cigar_gfa1_1_c_s = "9H8I7P6I5X4=3D2M1D"
  cigar_gfa1_1_rlen = 2+4+5+6+8
  cigar_gfa1_1_qlen = 1+2+3+4+5

  cigar_gfa2_1_s = "1M2I3D4P"
  cigar_gfa2_1_c_s = "4P3I2D1M"
  cigar_gfa2_1_rlen = 1+3
  cigar_gfa2_1_qlen = 1+2

  trace_1 = gfapy.Trace([12,12,0])
  trace_1_s = "12,12,0"

  cigar_invalid_value_1 = gfapy.CIGAR([
      gfapy.CIGAR.Operation(-12,"M"),
      gfapy.CIGAR.Operation(1,"D"),
      gfapy.CIGAR.Operation(2,"I")])
  cigar_invalid_value_2 = gfapy.CIGAR([
      gfapy.CIGAR.Operation(12, "Y"),
      gfapy.CIGAR.Operation(1,"D"),
      gfapy.CIGAR.Operation(2,"I")])
  cigar_invalid_type_1 = gfapy.CIGAR([
      "x",
      gfapy.CIGAR.Operation(1,"D"),
      gfapy.CIGAR.Operation(2,"I")])

  trace_invalid_value_1 = gfapy.Trace([-2,1,12])
  trace_invalid_type_1 = gfapy.Trace([12.0,1,12])

  cigar_empty = gfapy.CIGAR([])
  trace_empty = gfapy.Trace([])
  placeholder = gfapy.AlignmentPlaceholder()
  placeholder_s = "*"

  string_invalid = [
    "-12M1D2I", "12Y1D2I", "x1D2I",
    "-2,1,12", "12.0,1,12", "*x",
  ]

  cigar_op_1 = gfapy.CIGAR.Operation(1,"D")
  cigar_op_1_s = "1D"
  cigar_op_1_len = 1
  cigar_op_1_code = "D"
  cigar_op_2 = gfapy.CIGAR.Operation(2,"I")
  cigar_op_2_s = "2I"
  cigar_op_2_len = 2
  cigar_op_2_code = "I"

  def test_to_s(self):
    self.assertEqual(TestApiAlignment.cigar_1_s,     str(TestApiAlignment.cigar_1_s))
    self.assertEqual(TestApiAlignment.cigar_1_s,     str(TestApiAlignment.cigar_1))
    self.assertEqual(TestApiAlignment.trace_1_s,     str(TestApiAlignment.trace_1))
    self.assertEqual(TestApiAlignment.placeholder_s, str(TestApiAlignment.placeholder))
    self.assertEqual(TestApiAlignment.placeholder_s, str(TestApiAlignment.cigar_empty))
    self.assertEqual(TestApiAlignment.placeholder_s, str(TestApiAlignment.trace_empty))

  def test_cigar_clone(self):
    cigar1_clone = deepcopy(gfapy.Alignment(TestApiAlignment.cigar_1))
    self.assertEqual(TestApiAlignment.cigar_1_s, str(cigar1_clone))
    cigar1_clone[0].code = "="
    # copy is deep, only the clone has changed:
    self.assertNotEqual(TestApiAlignment.cigar_1_s, str(cigar1_clone))
    self.assertEqual(TestApiAlignment.cigar_1_s, str(TestApiAlignment.cigar_1))

  def test_to_alignment(self):
    self.assertEqual(TestApiAlignment.cigar_1,     gfapy.Alignment(TestApiAlignment.cigar_1_s))
    self.assertEqual(TestApiAlignment.trace_1,     gfapy.Alignment(TestApiAlignment.trace_1_s))
    self.assertEqual(TestApiAlignment.placeholder, gfapy.Alignment(TestApiAlignment.placeholder_s))
    for alignment in [TestApiAlignment.cigar_1, TestApiAlignment.trace_1, TestApiAlignment.cigar_empty,
        TestApiAlignment.trace_empty, TestApiAlignment.placeholder]:
      self.assertEqual(alignment, gfapy.Alignment(alignment))
    for string in TestApiAlignment.string_invalid:
      self.assertRaises(gfapy.FormatError, gfapy.Alignment, string)

  def test_decode_encode_invariant(self):
    for string in [TestApiAlignment.trace_1_s, TestApiAlignment.cigar_1_s, TestApiAlignment.placeholder_s]:
      self.assertEqual(string, str(gfapy.Alignment(string)))

  def test_is_placeholder(self):
    for a in [TestApiAlignment.cigar_empty, TestApiAlignment.trace_empty, TestApiAlignment.placeholder, TestApiAlignment.placeholder_s]:
      assert(gfapy.is_placeholder(a))
    for a in [TestApiAlignment.cigar_1, TestApiAlignment.cigar_1_s, TestApiAlignment.trace_1, TestApiAlignment.trace_1_s]:
      assert(not gfapy.is_placeholder(a))

  def test_validate(self):
    TestApiAlignment.trace_1.validate() # nothing raised
    TestApiAlignment.trace_empty.validate() # nothing raised
    TestApiAlignment.cigar_1.validate() # nothing raised
    TestApiAlignment.cigar_empty.validate() # nothing raised
    TestApiAlignment.placeholder.validate() # nothing raised
    self.assertRaises(gfapy.ValueError,TestApiAlignment.trace_invalid_value_1.validate)
    self.assertRaises(gfapy.ValueError,TestApiAlignment.cigar_invalid_value_1.validate)
    self.assertRaises(gfapy.ValueError,TestApiAlignment.cigar_invalid_value_2.validate)
    self.assertRaises(gfapy.TypeError,TestApiAlignment.trace_invalid_type_1.validate)
    self.assertRaises(gfapy.TypeError,TestApiAlignment.cigar_invalid_type_1.validate)

  def test_version_specific_validate(self):
    gfapy.Alignment(TestApiAlignment.cigar_gfa1_1_s,
        version="gfa1", valid=False) # nothing raised
    self.assertRaises(gfapy.FormatError, gfapy.Alignment,
        TestApiAlignment.cigar_gfa1_1_s, version="gfa2", valid=False)
    gfapy.Alignment(TestApiAlignment.cigar_gfa2_1_s,
        version="gfa1", valid=False) # nothing raised
    gfapy.Alignment(TestApiAlignment.cigar_gfa2_1_s,
        version="gfa2", valid=False) # nothing raised

  def test_array_methods(self):
    for a in [TestApiAlignment.cigar_empty, TestApiAlignment.trace_empty]:
      assert(not a)
    for a in [TestApiAlignment.cigar_1, TestApiAlignment.trace_1]:
      assert(a)
    self.assertEqual(gfapy.CIGAR.Operation(1,"D"), TestApiAlignment.cigar_1[1])
    self.assertEqual(12, TestApiAlignment.trace_1[1])


  def test_cigar_operation_methods(self):
    self.assertEqual(TestApiAlignment.cigar_op_1_len, TestApiAlignment.cigar_op_1.length)
    self.assertEqual(TestApiAlignment.cigar_op_1_code, TestApiAlignment.cigar_op_1.code)
    self.assertEqual(TestApiAlignment.cigar_op_1_s, str(TestApiAlignment.cigar_op_1))
    TestApiAlignment.cigar_op_1.length =  TestApiAlignment.cigar_op_2_len
    TestApiAlignment.cigar_op_1.code = TestApiAlignment.cigar_op_2_code
    self.assertEqual(TestApiAlignment.cigar_op_2, TestApiAlignment.cigar_op_1)
    self.assertEqual(TestApiAlignment.cigar_op_2_len, TestApiAlignment.cigar_op_1.length)
    self.assertEqual(TestApiAlignment.cigar_op_2_code, TestApiAlignment.cigar_op_1.code)
    self.assertEqual(TestApiAlignment.cigar_op_2_s, str(TestApiAlignment.cigar_op_2))

  def test_cigar_operation_validation(self):
    TestApiAlignment.cigar_op_1.validate() # nothing raised
    TestApiAlignment.cigar_op_1.validate(version="gfa2") # nothing raised
    TestApiAlignment.cigar_op_2.validate() # nothing raised
    TestApiAlignment.cigar_op_2.validate(version="gfa2") # nothing raised
    self.assertRaises(gfapy.VersionError, TestApiAlignment.cigar_op_1.validate, version="gfaX")
    stringlen = gfapy.CIGAR.Operation("1", "M")
    stringlen.validate() # nothing raised
    stringcode = gfapy.CIGAR.Operation(1, "M")
    stringcode.validate() # nothing raised
    malformed1 = gfapy.CIGAR.Operation([1], "M")
    self.assertRaises(gfapy.TypeError, malformed1.validate)
    malformed2 = gfapy.CIGAR.Operation(-1, "M")
    self.assertRaises(gfapy.ValueError, malformed2.validate)
    malformed3 = gfapy.CIGAR.Operation(1, "L")
    self.assertRaises(gfapy.ValueError, malformed3.validate)
    gfa1only = gfapy.CIGAR.Operation(1, "X")
    gfa1only.validate() # nothing raised
    self.assertRaises(gfapy.ValueError, gfa1only.validate, version="gfa2")

  def test_cigar_complement(self):
    self.assertEqual(TestApiAlignment.cigar_gfa1_1_c_s,
                 str(gfapy.Alignment(TestApiAlignment.cigar_gfa1_1_s, version="gfa1").complement()))
    self.assertEqual(TestApiAlignment.cigar_gfa2_1_c_s,
                 str(gfapy.Alignment(TestApiAlignment.cigar_gfa2_1_s).complement()))

  def test_cigar_length_on(self):
    self.assertEqual(TestApiAlignment.cigar_gfa1_1_rlen,
                 gfapy.Alignment(TestApiAlignment.cigar_gfa1_1_s,version="gfa1").
                 length_on_reference())
    self.assertEqual(TestApiAlignment.cigar_gfa1_1_qlen,
                 gfapy.Alignment(TestApiAlignment.cigar_gfa1_1_s,version="gfa1").
                 length_on_query())
    self.assertEqual(TestApiAlignment.cigar_gfa1_1_qlen,
                 gfapy.Alignment(TestApiAlignment.cigar_gfa1_1_c_s,version="gfa1").
                 length_on_reference())
    self.assertEqual(TestApiAlignment.cigar_gfa1_1_rlen,
                 gfapy.Alignment(TestApiAlignment.cigar_gfa1_1_c_s,version="gfa1").
                 length_on_query())
    self.assertEqual(TestApiAlignment.cigar_gfa2_1_rlen,
                 gfapy.Alignment(TestApiAlignment.cigar_gfa2_1_s).length_on_reference())
    self.assertEqual(TestApiAlignment.cigar_gfa2_1_qlen,
                 gfapy.Alignment(TestApiAlignment.cigar_gfa2_1_s).length_on_query())
    self.assertEqual(TestApiAlignment.cigar_gfa2_1_qlen,
                 gfapy.Alignment(TestApiAlignment.cigar_gfa2_1_c_s).length_on_reference())
    self.assertEqual(TestApiAlignment.cigar_gfa2_1_rlen,
                 gfapy.Alignment(TestApiAlignment.cigar_gfa2_1_c_s).length_on_query())




from .topology import Topology
from .references import References
from .validation import Validation
from .captured_path import CapturedPath
from .to_gfa2 import ToGFA2
from ..group import Group

class Path(Topology, References, Validation, CapturedPath, ToGFA2, Group):
  """A path line of a GFA1 file"""
  RECORD_TYPE = "P"
  POSFIELDS = ["path_name", "segment_names", "overlaps"]
  FIELD_ALIAS = { "name" : "path_name" }
  DATATYPE = {
    "path_name" : "path_name_gfa1",
    "segment_names" : "oriented_identifier_list_gfa1",
    "overlaps" : "alignment_list_gfa1",
  }
  NAME_FIELD = "path_name"
  REFERENCE_FIELDS = ["segment_names", "overlaps"]
  OTHER_REFERENCES = ["links"]

Path._apply_definitions()



"""
Encoding of python objects to GFA string representation
"""
import gfapy

class Writer:

  @staticmethod
  def _to_gfa_field(obj, datatype = None, safe = True, fieldname = None,
                   line = None):
    """Encode an object into its GFA string representation.

    The python object can be either an encoded GFA field (in which case it
    is already a string, thus it is only at most, depending on the other
    parameters, validated), or an object of a class compatible
    with the specified datatype, if a datatype is specified (see **datatype**),
    e.g. Integer # for i fields.

    Parameters
    ----------
    obj : object
      the python object to encode
    datatype : str
      datatype to use (one of `~gfapy.field.field.Field.FIELD_DATATYPE`);
      If none is specified, the datatype is used, which is returned by the
      `gfapy.field.Field._get_default_gfa_tag_datatype` method.
    fieldname : str, optional
      fieldname, for error messages
    line : gfapy.Line, optional
      line, for error messages
    safe : bool, optional
      *(defaults to: ***True***)* if **True**, the safe
      version of the encode function is used, which guarantees that the
      resulting data is valid; if **False**, the unsafe version is used,
      which, for some datatypes, skips validations in order to be faster
      than the safe version

    Raises
    ------
    gfapy.TypeError
      if an unknown datatype is specified, or the object type is not
      compatible with the datatype
    gfapy.ValueError
      if the object value is invalid for the datatype
    gfapy.FormatError
      if the object syntax is invalid for the
      datatype (eg for invalid encoded strings, if **safe** is set)
    """
    if not datatype:
      datatype = gfapy.Field._get_default_gfa_tag_datatype(obj)
    mod = gfapy.Field.FIELD_MODULE.get(datatype)
    if not mod:
      fieldnamemsg = "Field: {}\n".format(fieldname) if fieldname else ""
      contentmsg = "Content: {}\n".format(repr(obj))
      raise gfapy.TypeError(
        fieldnamemsg +
        contentmsg +
        "Datatype unknown: {}".format(repr(datatype)))
    try:
      if safe or not getattr(mod, "unsafe_encode"):
        return mod.encode(obj)
      else:
        return mod.unsafe_encode(obj)
    except Exception as err:
      fieldnamemsg = "Field: {}\n".format(fieldname) if fieldname else ""
      contentmsg = "Content: {}\n".format(repr(obj))
      datatypemsg = "Datatype: {}\n".format(datatype)
      raise err.__class__(
            fieldnamemsg +
            datatypemsg +
            contentmsg +
            str(err)) from err

  @staticmethod
  def _to_gfa_tag(obj, fieldname, datatype = None, line = None):
    """Representation of the data as a GFA tag.

    The representation is ``xx:d:content``, where ``xx`` is
    the tag name and ``d`` is the datatype.

    Parameters:
      obj (object): the python object to encode
      fieldname (string): the tag name
      datatype (string): (one of gfapy.Field.TAG_DATATYPE)
        the datatype; if not specified, the value returned by
        :func:``~gfapy.field.Field._get_default_gfa_tag_datatype``
        is used.
      line (string): the line content, for error messages
    """
    if not datatype:
      datatype = gfapy.Field._get_default_gfa_tag_datatype(obj)
    return "{}:{}:{}".format(fieldname, datatype,
            Writer._to_gfa_field(obj, datatype = datatype,
              fieldname = fieldname, line = line))



import unittest
import gfapy

class TestUnitOrientedLine(unittest.TestCase):

  def test_init(self):
    a = gfapy.OrientedLine("a","+")
    # no validation on creation: (invalid orientation)
    gfapy.OrientedLine("a","*")
    # no validation on creation: (invalid line name)
    gfapy.OrientedLine("a\ta","+")
    b = gfapy.OrientedLine("a+")
    self.assertEqual(a, b)
    c = gfapy.OrientedLine(["a","+"])
    self.assertEqual(a, c)
    self.assertRaises(IndexError, gfapy.OrientedLine, [])
    self.assertRaises(IndexError, gfapy.OrientedLine, ["a"])
    # nothing raised, if too many args are provided (further are ignored)
    gfapy.OrientedLine(["a", "+", 1])

  def test_properties(self):
    a = gfapy.OrientedLine("a", "+")
    self.assertEqual("a", a.line)
    self.assertEqual("+", a.orient)
    self.assertEqual("a", a.name)
    s = gfapy.Line("S\tb\t*\txx:Z:1.0")
    a.line = s
    self.assertEqual(s, a.line)
    self.assertEqual("b", a.name)
    self.assertEqual("+", a.orient)
    a.orient = "-"
    self.assertEqual(s, a.line)
    self.assertEqual("-", a.orient)

  def test_validate(self):
    gfapy.OrientedLine("a","+").validate()
    gfapy.OrientedLine(gfapy.Line("S\tb\t*\txx:Z:1.0"),
                       "-").validate()
    self.assertRaises(gfapy.ValueError,
        gfapy.OrientedLine("a","*").validate)
    self.assertRaises(gfapy.TypeError,
        gfapy.OrientedLine([],"+").validate)
    self.assertRaises(gfapy.FormatError,
        gfapy.OrientedLine("a\ta","+").validate)


  def test_inverted(self):
    os = gfapy.OrientedLine("a", "+")
    inv_os = os.inverted()
    self.assertEqual("a", inv_os.line)
    self.assertEqual("+", os.orient)
    self.assertEqual("-", inv_os.orient)
    s = gfapy.Line("S\tb\t*\txx:Z:1.0")
    os = gfapy.OrientedLine(s, "-")
    inv_os = os.inverted()
    self.assertEqual(s, inv_os.line)
    self.assertEqual("-", os.orient)
    self.assertEqual("+", inv_os.orient)
    os = gfapy.OrientedLine("a", "*")
    self.assertRaises(gfapy.ValueError, os.invert)

  def test_str(self):
    self.assertEqual("a-", str(gfapy.OrientedLine("a","-")))
    s = gfapy.Line("S\tb\t*\txx:Z:1.0")
    self.assertEqual("b+", str(gfapy.OrientedLine(s,"+")))

  def test_equal(self):
    a = gfapy.OrientedLine("a", "+")
    b = gfapy.OrientedLine(gfapy.Line("S\ta\t*"), "+")
    c = gfapy.OrientedLine("a", "-")
    self.assertEqual(a, b)
    self.assertNotEqual(a, c)
    # line itself is not checked for equiv, only name:
    b2 = gfapy.OrientedLine(gfapy.Line("S\ta\tCACAC"), "+")
    self.assertEqual(b, b2)
    # equivalence to string:
    self.assertEqual("a+", a)
    self.assertEqual("a+", b)
    self.assertEqual(a, "a+")
    self.assertEqual(b, "a+")
    # equivalence to list:
    self.assertEqual(a, ["a", "+"])
    self.assertEqual(b, ["a", "+"])
    self.assertEqual(["a", "+"], a)
    self.assertEqual(["a", "+"], b)

  def test_block(self):
    a = gfapy.OrientedLine("a", "+")
    a._block()
    with self.assertRaises(gfapy.RuntimeError):
      a.line = "b"
    a._unblock()
    a.line = "b"

  def test_delegate_methods(self):
    ol = gfapy.OrientedLine(gfapy.Line("S\ta\tCACAC"), "+")
    self.assertEqual("CACAC", ol.sequence)
    self.assertEqual("CACAC", ol.field_to_s("sequence"))
    ol.set("xx", 1)
    self.assertEqual("xx:i:1", ol.field_to_s("xx", True))
